"""
Module otsun.source that implements rays and its sources
"""

import itertools

import Part
import numpy as np
from FreeCAD import Base

from .math import pick_random_from_cdf, myrandom, tabulated_function, two_orthogonal_vectors, area_of_triangle, random_point_of_triangle
from .optics import dispersion_from_main_direction, random_polarization, dispersion_polarization
from .ray import Ray
from random import choices

from scipy.spatial import ConvexHull

EPSILON = 1E-6
# Tolerance for considering equal to zero

class GeneralizedSunWindow(object):
    def __init__(self, scene, main_direction):
        bbs = []
        for shape in itertools.chain(scene.solids, scene.faces):
            bbs.append(shape.BoundBox)
        origin = (scene.boundbox.Center -
                          main_direction * 0.5 * scene.boundbox.DiagonalLength)
        u, v = two_orthogonal_vectors(main_direction)
        points = []
        plane_points = []
        for bb in bbs:
            xs = [bb.XMin, bb.XMax]
            ys = [bb.YMin, bb.YMax]
            zs = [bb.ZMin, bb.ZMax]
            coords = itertools.product(xs, ys, zs)
            for c in coords:
                point = Base.Vector(c)
                points.append(point)
                pu = u.dot(point - origin) * 1.02
                pv = v.dot(point - origin) * 1.02
                plane_points.append([pu,pv])
        hull = ConvexHull(plane_points)
        plane_vertices = [plane_points[i] for i in hull.vertices]
        self.vertices = [origin + u*pu + v*pv for (pu,pv) in plane_vertices]
        self.triangles = [[self.vertices[0], self.vertices[i], self.vertices[i+1]]
                          for i in range(1,len(self.vertices)-1)]
        self.triangle_areas = list(map(area_of_triangle, self.triangles))
        self.aperture = sum(self.triangle_areas)
        self.main_direction = main_direction

    def add_to_document(self, doc):
        sw = Part.makePolygon(self.vertices, True)
        doc.addObject("Part::Feature", "SunWindow").Shape = sw

    def random_point(self):
        random_triangle = choices(self.triangles, self.triangle_areas)[0]
        return random_point_of_triangle(random_triangle)

    def random_direction(self):
        """
        Returns the main direction

        Maybe in the future will return some random vector

        Returns
        -------
        Base.Vector
        """
        return self.main_direction





class SunWindow(object):
    """
    Class that implements a Sun window (rectangle that emits rays)

    Source of light defined by a rectangle "at infinity" in space that emits rays perpendicular to it.

    Parameters
    ----------
    scene : Scene
        Scene that contains the sun window
    main_direction : Base.Vector
        Vector orthogonal to the emitting region

    Attributes
    ----------
    origin : Base.Vector
        Center of the rectangle
    v1, v2 : Base.Vector
        Unit vectors parallel to the sides of the rectangle
    length1, length2 : float
        Lengths of the sides of the rectangle
    aperture : float
        Area of the rectangle
    """

    def __init__(self, scene, main_direction):
        bbs = []
        for shape in itertools.chain(scene.solids, scene.faces):
            bbs.append(shape.BoundBox)
        projected_points = []
        for bb in bbs:
            xs = [bb.XMin, bb.XMax]
            ys = [bb.YMin, bb.YMax]
            zs = [bb.ZMin, bb.ZMax]
            coords = itertools.product(xs, ys, zs)
            points = [Base.Vector(c) for c in coords]

            point_of_plane = (scene.boundbox.Center -
                              main_direction * 0.5 * scene.boundbox.DiagonalLength)
            projected_points.extend([p.projectToPlane(point_of_plane, main_direction)
                                for p in points])
        (self.origin, self.v1, self.v2, self.length1, self.length2) = (
            SunWindow.find_min_rectangle(projected_points, main_direction))
        self.aperture = self.length1 * self.length2
        self.main_direction = main_direction

    @staticmethod
    def find_min_rectangle(points, normal):
        """
        Computes the minimum rectangle covering points in a direction

        Given a list of `points`, take its projection in a `normal` direction,
        and the rectangle with minimum area that encloses this projections

        Parameters
        ----------
        points : list of Base.Vector
            List of points to enclose
        normal : Base.Vector
            Vector orthogonal to the rectangle to be found
        Returns
        -------
        origin : Base.Vector
            Center of the rectangle
        best_v1, best_v2 : Base.Vector
            Unit vector parallel to the sides of the rectangle
        best_v2 : Base.Vector
            Other unit vector parallel to the sides of the rectangle
        length1 : float
            Length of side of the rectangle
        length2 : float
            Length of other side of the rectangle
        """
        min_area = None
        for (p, q) in itertools.combinations(points, 2):
            v1 = q - p
            if v1.Length < EPSILON:  # TODO: customize
                continue
            v1.normalize()
            v2 = v1.cross(normal)
            v2.normalize()
            xs = [v1.dot(r - p) for r in points]
            ys = [v2.dot(r - p) for r in points]
            minx = min(xs)
            maxx = max(xs)
            miny = min(ys)
            maxy = max(ys)
            length1 = maxx - minx
            length2 = maxy - miny
            area = length1 * length2
            if not min_area or area < min_area:
                min_area = area
                best_origin = p + v1 * minx + v2 * miny
                best_v1 = v1
                best_v2 = v2
                best_length1 = length1
                best_length2 = length2
        length1 = best_length1 * 1.04
        length2 = best_length2 * 1.04
        origin = best_origin - best_v1 * length1 * 0.02 - best_v2 * length2 * 0.02
        return origin, best_v1, best_v2, length1, length2

    def random_point(self):
        """
        Returns a random point on the rectangle

        Returns
        -------
        Base.Vector
        """
        return (self.origin + self.v1 * self.length1 * myrandom() +
                self.v2 * self.length2 * myrandom())

    def random_direction(self):
        """
        Returns the main direction

        Maybe in the future will return some random vector

        Returns
        -------
        Base.Vector
        """
        return self.main_direction

    def add_to_document(self, doc):
        """
        Adds the rectangle to the FreeCAD document

        Parameters
        ----------
        doc : App.Document
        """
        sw = Part.makePolygon([self.origin,
                               self.origin + self.v1 * self.length1,
                               self.origin + self.v1 * self.length1 +
                               self.v2 * self.length2,
                               self.origin + self.v2 * self.length2
                               ], True)
        doc.addObject("Part::Feature", "SunWindow").Shape = sw


class LightSource(object):
    """
    Sets up a light source with a given scene, a given emitting region and a given light spectrum.
    The emitting region must provide the main direction.
    Light spectrum could be: a constant value (for a single wavelength in nanometers), or a spectrum distribution.
    The distribution (dispersion) for the main direction is provided in "direction_distribution".
    The polarization_vector is a Base.Vector for polarized light. If is not given unpolarized light is generated.
    """

    def __init__(self, scene, emitting_region, light_spectrum, initial_energy, direction_distribution=None,
                 polarization_vector=None):
        self.scene = scene
        self.emitting_region = emitting_region
        self.light_spectrum = light_spectrum
        self.initial_energy = initial_energy
        self.direction_distribution = direction_distribution
        self.polarization_vector = polarization_vector
        self.wavelengths = []

    def emit_ray(self):
        """
        Simulates the emission of a ray
        """
        point = self.emitting_region.random_point()
        main_direction = self.emitting_region.main_direction  # emitting main direction
        direction = main_direction
        if self.direction_distribution is not None:  # main direction has a distribution
            theta = self.direction_distribution(myrandom())
            phi = 360.0 * myrandom()
            direction = dispersion_from_main_direction(main_direction, theta, phi)
            if self.polarization_vector:  # single polarization vector is active
                polarization_vector = dispersion_polarization(main_direction, self.polarization_vector, theta, phi)
        if self.polarization_vector is None:  # unpolarization is active
            polarization_vector = random_polarization(direction)  # random polarization from light direction
        else:
            polarization_vector = self.polarization_vector
            polarization_vector.normalize()
        if np.isscalar(self.light_spectrum):
            wavelength = self.light_spectrum  # experiment with a single wavelength (nanometers)
        else:
            wavelength = pick_random_from_cdf(self.light_spectrum)  # light spectrum is active (nanometers)
        ray = Ray(self.scene, point, direction, wavelength, self.initial_energy, polarization_vector)
        return ray


# Auxiliary functions for buie_distribution

def _calculate_a1(CSR, SD):
    """ Parameter a1 needed for the normalization of the probability distribution in the disk region
    """

    th = np.arange(0.0, SD, 0.001)
    values = [2.0 * np.pi * np.cos(0.326 * angle) / np.cos(0.305 * angle) * angle
              for angle in th]
    a1 = (1.0 - CSR) / np.trapz(values, dx=0.001)
    return a1


def _circumsolar__density_distribution(angle, CSR):
    gamma = 2.2 * np.log(0.52 * CSR) * CSR ** 0.43 - 0.1
    kappa = 0.9 * np.log(13.5 * CSR) * CSR ** (-0.3)
    return 2.0 * np.pi * np.exp(kappa) * angle ** (gamma + 1.0)


def _calculate_a2(CSR, SD, SS):
    """ Parameter a2 needed for the normalization of the probability distribution in the circumsolar region"""
    f = _circumsolar__density_distribution
    th = np.arange(SD, SS, 0.001)
    f_th = np.vectorize(f)
    a2 = CSR / np.trapz(f_th(th, CSR), dx=0.001)
    return a2


def _calculate_CDF_disk_region(a1, SD):
    """
    Cumulative Distribution Function in the solar disk region
    """
    th = np.arange(0.0, SD, 0.001)
    values = [2.0 * np.pi * np.cos(0.326 * angle) / np.cos(0.305 * angle) * angle
              for angle in th]
    cumulative = np.cumsum(values)
    CDF = (th, a1 * cumulative / 1000.0)
    return CDF


def _th_solar_disk_region(u, CDF):
    """ Random angle based on the probability distribution in the circumsolar region"""
    # TODO: It is recomputed every time? @Ramon
    # CDF = CDF_Disk_Region # calculate_CDF_disk_region(a1, SD)
    idx = (np.abs(CDF[1] - u)).argmin()
    return CDF[0][idx]


def _th_circumsolar_region(u, CSR, SD, a2):
    """ Random angle based on the CDF in the circumsolar region"""
    gamma = 2.2 * np.log(0.52 * CSR) * CSR ** (0.43) - 0.1
    kappa = 0.9 * np.log(13.5 * CSR) * CSR ** (-0.3)
    u_csr = (u - 1.0) + CSR  # Since CSR-CDF starts at zero
    f1 = u_csr * (gamma + 2.0) / (a2 * 2 * np.pi * np.exp(kappa))
    f2 = SD ** (gamma + 2.0)
    exp = (1.0 / (gamma + 2.0))
    th_u = np.power(f1 + f2, exp)
    return th_u


def buie_distribution(CircumSolarRatio):
    """
    Implementation of the Buie Distribution for Sun emission

    A distribution for the direct light from the sun according to the Buie model:
    Buie, D., 2005. Corrigendum to "The effective size of the solar cone for
    solar concentrating systems" [Solar Energy 74 (2003) 417-427].
    Solar Energy 79, 568-570. 5.2

    Parameters
    ----------
    CircumSolarRatio : float

    Returns
    -------
    angle distribution for random input: function
        Function that interpolates by straight line segments the input data
    """
    CSR = CircumSolarRatio
    SD = 4.65
    # Solar Disk in mrad
    SS = 43.6
    # Solar Size in mrad
    a2 = None # _calculate_a2(CSR, SD, SS)
    CDF_Disk_Region = None # _calculate_CDF_disk_region(a1, SD)
    #    Buie distribution for the disk region
    u_values = np.arange(0.0, 1.001, 0.001)
    dist_values = []
    for u in u_values:
        if u <= 1.0 - CSR:
            if CDF_Disk_Region is None:
                a1 = _calculate_a1(CSR, SD)
                #     normalization constant for the disk region
                CDF_Disk_Region = _calculate_CDF_disk_region(a1, SD)
            dist_values.append(_th_solar_disk_region(u, CDF_Disk_Region) / 1000.0 * 180.0 / np.pi)
        else:
            if a2 is None:
                a2 = _calculate_a2(CSR, SD, SS)
                #    normalization constant for the circumsolar region
            dist_values.append(_th_circumsolar_region(u, CSR, SD, a2) / 1000.0 * 180.0 / np.pi)
    f = tabulated_function(u_values, dist_values)
    return f



"""Module otsun.outputs

Helper functions to format data for output
"""

import numpy as np

def spectrum_to_constant_step(file_in, wavelength_step, wavelength_min, wavelength_max):
    data_array = np.loadtxt(file_in, usecols=(0, 1))
    wl_spectrum = data_array[:, 0]
    I_spectrum = data_array[:, 1]
    array_inter = [[x, np.interp(x, wl_spectrum, I_spectrum)] for x in
                   np.arange(wavelength_min, wavelength_max + wavelength_step / 2.0, wavelength_step)]
    return np.array(array_inter)


def make_histogram_from_experiment_results(results_wavelength, results_energy, step_wavelength, aperture_collector,
                                           aperture_source):
    y_energy = np.array(np.concatenate(results_energy))
    y_energy = (y_energy / aperture_collector) / (1.0 / aperture_source)
    x_wavelength = np.array(np.concatenate(results_wavelength))
    data_ = np.array([x_wavelength, y_energy])
    data_ = data_.T
    bins_ = np.arange(int(np.amin(x_wavelength)), np.amax(x_wavelength) + step_wavelength * 1.1, step_wavelength)
    table_ = np.histogram(data_[:, 0], bins=bins_, weights=data_[:, 1])
    norm = np.histogram(data_[:, 0], bins=bins_)
    u = np.divide(table_[0], norm[0])
    bins = np.arange(int(np.amin(x_wavelength)), np.amax(x_wavelength) + step_wavelength, step_wavelength)
    table_ = np.column_stack((bins, u))
    return table_


def twoD_array_to_constant_step(twoD_array, step, wavelength_min, wavelength_max):
    wl_spectrum = twoD_array[:, 0]
    I_spectrum = twoD_array[:, 1]
    array_inter = [[x, np.interp(x, wl_spectrum, I_spectrum)] for x in
                   np.arange(wavelength_min, wavelength_max + step / 2.0, step)]
    return np.array(array_inter)


def spectral_response(optical_absorption_wavelength, iqe):
    q_e = 1.60217662E-19
    h = 6.62607E-34
    c = 299792458.0
    hc = h * c
    opt_wavelength = optical_absorption_wavelength
    if np.isreal(iqe):
        SR = [[opt[0], iqe * opt[0] * opt[1] * q_e * 1E-9 / hc, ] for opt in opt_wavelength]
    else:
        data_array = np.loadtxt(iqe, usecols=(0, 1))
        wl_ = data_array[:, 0]
        iqe_ = data_array[:, 1]
        SR = [[opt[0], np.interp(opt[0], wl_, iqe_) * opt[0] * opt[1] * q_e * 1E-9 / hc, ] for opt in opt_wavelength]
    return np.array(SR)


def photo_current(spectral_response, source_spectrum):
    wavelengths = source_spectrum[:, 0]
    SR_by_spectrum = spectral_response[:, 1] * source_spectrum[:, 1]
    photo_current = np.trapz(SR_by_spectrum, x=wavelengths)
    return photo_current


# ---
# Helper functions for outputs in Total Analysis
# ---

def integral_from_data_file(file_in):
    source_spectrum = np.loadtxt(file_in, usecols=(0, 1))
    integral = np.trapz(source_spectrum[:, 1], x=source_spectrum[:, 0])
    return integral



"""Module otsun.materials for treating materials

The module relies on a basic class `Material` with two subclasses
`VolumeMaterial` and `SurfaceMaterial`, and several subclasses of them
for specific materials.
"""

import json
import zipfile
from FreeCAD import Base
from .optics import Phenomenon, OpticalState, refraction, matrix_reflectance, \
    calculate_reflectance, shure_refraction, reflection_hub, reflection, outgoing_polarization
from .math import arccos, parallel_orthogonal_components, rad_to_deg, myrandom, normalize, \
    constant_function, correct_normal, tabulated_function
from numpy import sqrt
import numpy as np
from autologging import traced
import logging
logger = logging.getLogger(__name__)
import pandas as pd


class NumpyEncoder(json.JSONEncoder):
    """Wrapper to dump numpy arrays as json"""

    def default(self, obj):
        if isinstance(obj, np.ndarray):
            return obj.tolist()
        return json.JSONEncoder.default(self, obj)


def load_from_txt_or_csv(file):
    df = pd.read_table(
        file,
        sep=None,
        index_col=0,
        header=None,
        comment='#',
        engine='python'
    ).interpolate(method='index')
    df.insert(0, 0, df.index)
    return df.values

class MaterialNotFoundError(Exception):
    pass

@traced(logger)
class Material(object):
    """
    Class used to represent materials and their physical properties

    The `properties` dictionary holds the physical properties of the material.
    Its contents are specific to the kind of material.

    Parameters
    ----------
    name : str
        Name of the material
    properties : dict
        Dictionary with physical properties of the material

    Attributes
    ----------
    classname : str
        String identifying the class
    """

    by_name = {}
    """
    Dict that associates the name of each created material with the material itself
    """

    def __init__(self, name, properties=None):
        self.by_name[name] = self
        self.name = name
        self.classname = ""
        if properties is None:
            properties = {}
        self.properties = properties

    @staticmethod
    def plain_properties_to_properties(plain_properties):
        """
        Converts properties of a material in plain format (json) to internal format

        Parameters
        ----------
        plain_properties : dict

        Returns
        -------
        dict
        """
        properties = {}
        for key in plain_properties:
            plain_property = plain_properties[key]
            prop_type = plain_property['type']
            prop_value = plain_property['value']
            if prop_type == 'scalar':
                properties[key] = prop_value
            if prop_type == 'constant':
                properties[key] = constant_function(prop_value)
            if prop_type == 'tabulated':
                properties[key] = tabulated_function(
                    np.array(prop_value[0]), np.array(prop_value[1]))
            if prop_type == 'matrix':
                properties[key] = matrix_reflectance(np.array(prop_value))
        properties['plain_properties'] = plain_properties
        return properties

    @staticmethod
    def properties_to_plain_properties(properties):
        """
        Converts properties of a material in internal format to plain (json ready) format

        Since the plain properties are stored in the internal format,
        no need for conversions

        Parameters
        ----------
        properties : dict

        Returns
        -------
        dict
        """
        return properties.get('plain_properties', None)

    @classmethod
    def get_from_label(cls, label):
        """
        Returns the material given its label

        Given a `label` of an object (from a FreeCad document) of the form
        XXX(MatYYY,OTHER_INFO), returns the material whose name is MatYYY

        Parameters
        ----------
        label : str

        Returns
        -------
        Material
        """
        if ("(" not in label) or (")" not in label):
            return None
        start = label.find("(")
        end = label.find(")")
        string = label[start + 1:end]
        name = string.split(',')[0]
        if name in cls.by_name:
            return cls.by_name[name]
        else:
            raise MaterialNotFoundError(f"Material {name} not found")

    @classmethod
    def create(cls, name, properties):
        """Wrapper to create a material"""
        _ = cls(name, properties)

    @classmethod
    def load_from_json(cls, info):
        if type(info).__name__ == 'dict':
            info = [info]
        names = []
        for mat_spec in info:
            classname = mat_spec['classname']
            logger.debug(classname)
            the_class = globals()[classname]
            name = mat_spec['name']
            names.append(name)
            if issubclass(the_class, TwoLayerMaterial):
                name_front_layer = mat_spec['name_front_layer']
                name_back_layer = mat_spec['name_back_layer']
                mat = TwoLayerMaterial(name, name_front_layer, name_back_layer)
            else:
                plain_properties = mat_spec['plain_properties']
                properties = the_class.plain_properties_to_properties(plain_properties)
                mat = Material(name, properties)
            mat.__class__ = the_class
        if len(names) == 1:
            return names[0]
        else:
            return names

    @classmethod
    def load_from_json_fileobject(cls, f):
        """
        Load materials from a json fileobject

        If the file contains a single dict, then it means that it contains a
        single material. Otherwise it contains an array, where each entry is a dict
        representing a material.

        Parameters
        ----------
        f : BinaryIO
            File object

        Returns
        -------
        str
            String with the name of the last material imported from the file
        """
        info = json.load(f)
        return cls.load_from_json(info)

    @classmethod
    def load_from_json_file(cls, filename):
        """
        Load materials from a json file

        Parameters
        ----------
        filename : str
            Name of the file

        Returns
        -------
        str
            String with the name of the last material imported from the file
        """
        try:
            with open(filename, 'rb') as f:
                return cls.load_from_json_fileobject(f)
        except IOError:
            logger.exception("error in processing file %s", filename)

    @classmethod
    def load_from_json_zip(cls, filename):
        """
        Load all materials from a zip file
        Parameters
        ----------
        filename : str
            Name of the file
        """
        try:
            with zipfile.ZipFile(filename) as z:
                for matfile in z.namelist():
                    try:
                        with z.open(matfile) as f:
                            cls.load_from_json_fileobject(f)
                    except:
                        logger.exception("File %s in zip contains errors", matfile)
        except IOError:
            logger.exception("error in processing file %s", filename)

    def get_n(self, wavelength):
        """
        Returns the (complex) refractive index at a certain wavelength

        Parameters
        ----------
        wavelength : float

        Returns
        -------
            complex
        """
        n = self.properties['index_of_refraction'](wavelength)
        if 'extinction_coefficient' in self.properties:
            kappaf = self.properties['extinction_coefficient']
            kappa = kappaf(wavelength)
            return n + 1j * kappa
        else:
            return n

    def change_of_optical_state(self, *args):
        """
        Computes how a ray behaves when interacting with the material.
        MUST be subclassed

        The material where the ray is actually located is held in
        ray.current_material.

        Parameters
        ----------
        args
            Variable length argument list

        Returns
        -------
        OpticalState
        """
        pass

    def to_json(self):
        """Converts material to json. MUST be subclassed"""
        return ""

    def save_to_json_file(self, filename):
        """
        Save material to json file

        Parameters
        ----------
        filename : str
            Name of the file
        """
        with open(filename, 'w') as f:
            f.write(self.to_json())

    @staticmethod
    def all_to_json():
        """
        Convert all materials to json

        Returns
        -------
        list of str
        """
        materials = Material.by_name.values()
        simple_mats = [material for material in materials if
                       not isinstance(material, TwoLayerMaterial)]
        composite_mats = [material for material in materials if
                          isinstance(material, TwoLayerMaterial)]
        all_mats = simple_mats + composite_mats
        return [mat.to_json() for mat in all_mats]

    @staticmethod
    def save_all_to_json_file(filename):
        """
        Saves all materials to text file

        Parameters
        ----------
        filename
        """
        with open(filename, 'w') as f:
            f.write('[\n')
            f.write(',\n'.join(Material.all_to_json()))
            f.write('\n]')

    def get_lambertian_weight_and_kind(self):
        if 'lambertian_material' in self.properties:
            # For backwards compatibility. In a previous version only "pure" lambertian
            # with total dispersion was considered
            if self.properties['lambertian_material']:
                weight, kind = 1, "Total"
            else:
                weight, kind = 0, None
        else:
            weight = self.properties.get('lambertian_weight', 0)
            kind = self.properties.get('lambertian_kind', None)
        return weight, kind


@traced(logger)
class VolumeMaterial(Material):
    """
    Subclass of Material for materials with volume

    The `properties` parameter must be a dict with the physical properties
    describing the material. At least, the following must be provided:
    'index_of_refraction': float (index of refraction of the material, as a function
    of its wavelength, only real part)
    'extinction_coefficient': float (imaginary part of the index of refraction
    of the material as a function of its wavelength)
    """

    def __init__(self, name, properties=None):
        super(VolumeMaterial, self).__init__(name, properties)

    def change_of_optical_state(self, ray, normal_vector):
        """
        Compute the change of optical state

        Computes the new optical state when `ray` hits the material
        at a point with given `normal_vector`

        Parameters
        ----------
        ray : Ray
        normal_vector : Base.Vector

        Returns
        -------
        OpticalState
        """
        wavelength = ray.wavelength
        if isinstance(ray.current_medium(), PolarizedThinFilm):
            return OpticalState(ray.current_polarization(),
                                ray.current_direction(), Phenomenon.REFRACTION, self)  # TODO: Set solid
        else:
            n1 = ray.current_medium().get_n(wavelength)
            n2 = self.get_n(wavelength)
            optical_state = refraction(ray.current_direction(), normal_vector, n1, n2, ray.current_polarization())
            if optical_state.phenomenon == Phenomenon.REFRACTION:
                optical_state.material = self  # TODO: Set solid
            else:
                optical_state.material = ray.current_medium()  # TODO: Set solid
            return optical_state

    def to_json(self):
        """
        Dumps the material in json format.
        """
        return json.dumps(
            {
                'name': self.name,
                'classname': self.__class__.__name__,
                'plain_properties': self.properties.get(
                    'plain_properties', None)
            }, cls=NumpyEncoder, indent=4
        )


class SimpleVolumeMaterial(VolumeMaterial):
    """
    Subclass of `VolumeMaterial` for those materials with constant properties.
    """

    def __init__(self, name, index_of_refraction, attenuation_coefficient=None):
        plain_properties = {
            'index_of_refraction': {
                'type': 'constant',
                'value': index_of_refraction
            },
            'attenuation_coefficient': {
                'type': 'constant',
                'value': attenuation_coefficient
            }
        }
        super(SimpleVolumeMaterial, self).__init__(name, {})
        self.properties = Material.plain_properties_to_properties(plain_properties)


class WavelengthVolumeMaterial(VolumeMaterial):
    """
    Subclass of `VolumeMaterial` for materials with tabulated index of refraction.
    """

    def __init__(self, name, file_index_of_refraction):
        data_refraction = load_from_txt_or_csv(file_index_of_refraction)
        # data_refraction = np.loadtxt(file_index_of_refraction, usecols=(0, 1, 2))
        wavelength_values = data_refraction[:, 0]
        n_values = data_refraction[:, 1]
        k_values = data_refraction[:, 2]
        plain_properties = {
            'index_of_refraction': {
                'type': 'tabulated',
                'value': [wavelength_values, n_values]
            },
            'extinction_coefficient': {
                'type': 'tabulated',
                'value': [wavelength_values, k_values]
            }
        }
        super(WavelengthVolumeMaterial, self).__init__(name)
        self.properties = Material.plain_properties_to_properties(plain_properties)


class PVMaterial(VolumeMaterial):
    """
    Subclass of `VolumeMaterial` for photovoltaic materials.
    """

    def __init__(self, name, file_index_of_refraction):
        # file_index_of_refraction with three columns:
        # wavelenth in nm,
        # real(index of refraction),
        # imaginary(index of refraction)
        data_refraction = load_from_txt_or_csv(file_index_of_refraction)
        # data_refraction = np.loadtxt(file_index_of_refraction, usecols=(0, 1, 2))
        wavelength_values = data_refraction[:, 0]
        n_values = data_refraction[:, 1]
        k_values = data_refraction[:, 2]
        plain_properties = {
            'index_of_refraction': {
                'type': 'tabulated',
                'value': [wavelength_values, n_values]
            },
            'extinction_coefficient': {
                'type': 'tabulated',
                'value': [wavelength_values, k_values]
            },
            'PV_material': {
                'type': 'scalar',
                'value': True
            }
        }
        super(PVMaterial, self).__init__(name)
        self.properties = Material.plain_properties_to_properties(plain_properties)

    def get_PV_data(self, ray, energy_before):
        """
        Computes the photovoltaic data stored in a ray.

        Parameters
        ----------
        ray : Ray
            Ray that has passed through the PV material
        energy_before : float
            Energy of the ray before passing through the PV material

        Returns
        -------
        float, tuple of floats
        """
        alpha = self.properties['extinction_coefficient'](
            ray.wavelength) * 4 * np.pi / (ray.wavelength / 1E6)  # mm-1
        # TODO: @Ramon: Revisar angle (sona raro)
        angle_incident = arccos(
            - ray.last_normal.dot(ray.current_direction())) * 180.0 / np.pi
        point_1 = ray.points[-1]
        point_2 = ray.points[-2]
        return (energy_before - ray.energy,
                (point_2.x, point_2.y, point_2.z, point_1.x, point_1.y, point_1.z,
                 energy_before, ray.energy, ray.wavelength, alpha, angle_incident, self.name)
                )


class PolarizedThinFilm(VolumeMaterial):
    """
    Subclass of `VolumeMaterial` for polarized thin film materials.
    """

    def __init__(self, name, file_thin_film, file_front, file_back):
        # thin film material calculated by TMM method, six columns:
        # wavelenth in nm,
        # angle in deg.,
        # reflectance s-polarized (perpendicular),
        # reflectance p-polarized (parallel),
        # transmittance s-polarized,
        # transmittance p-polarized
        # the values in coating_material should be in the corresponding
        # order columns
        ### data = np.loadtxt(file_thin_film)
        data = load_from_txt_or_csv(file_thin_film)
        data_reflectance = data[:, [0, 1, 2, 3]]
        data_transmittance = data[:, [0, 1, 4, 5]]
        if file_front != 'Vacuum':
            data_refraction_front = load_from_txt_or_csv(file_front)
            # data_refraction_front = np.loadtxt(file_front, usecols=(0, 1, 2))
            wavelength_values_front = data_refraction_front[:, 0]
            n_values_front = data_refraction_front[:, 1]
            k_values_front = data_refraction_front[:, 2]
            index_of_refraction_front = {
                'type': 'tabulated',
                'value': [wavelength_values_front, n_values_front]
            }
            extinction_coefficient_front = {
                'type': 'tabulated',
                'value': [wavelength_values_front, k_values_front]
            }
        else:
            index_of_refraction_front = {
                'type': 'constant',
                'value': 1.0
            }
            extinction_coefficient_front = {
                'type': 'constant',
                'value': 0.0
            }
        if file_back != 'Vacuum':
            data_refraction_back = load_from_txt_or_csv(file_back)
            # data_refraction_back = np.loadtxt(file_back, usecols=(0, 1, 2))
            wavelength_values_back = data_refraction_back[:, 0]
            n_values_back = data_refraction_back[:, 1]
            k_values_back = data_refraction_back[:, 2]
            index_of_refraction_back = {
                'type': 'tabulated',
                'value': [wavelength_values_back, n_values_back]
            }
            extinction_coefficient_back = {
                'type': 'tabulated',
                'value': [wavelength_values_back, k_values_back]
            }
        else:
            index_of_refraction_back = {
                'type': 'constant',
                'value': 1.0
            }
            extinction_coefficient_back = {
                'type': 'constant',
                'value': 0.0
            }
        plain_properties = {
            'Matrix_reflectance_thin_film': {
                'type': 'matrix',
                'value': data_reflectance
            },
            'Matrix_transmittance_thin_film': {
                'type': 'matrix',
                'value': data_transmittance
            },
            'index_of_refraction_front': index_of_refraction_front,
            'extinction_coefficient_front': extinction_coefficient_front,
            'index_of_refraction_back': index_of_refraction_back,
            'extinction_coefficient_back': extinction_coefficient_back,
            'thin_film': {
                'type': 'scalar',
                'value': True
            },
        }
        super(PolarizedThinFilm, self).__init__(name)
        self.properties = Material.plain_properties_to_properties(plain_properties)

    @staticmethod
    def calculate_state_thin_film(incident, normal_vector, n1, n2, polarization_vector,
                                  properties, wavelength):
        """
        Helper function for the computation of the optical state once the ray has passed through the film
        """
        # returns optical state of the ray in thin film material
        normal = correct_normal(normal_vector, incident)
        backside = False
        if normal != normal_vector:
            backside = True
        r = n1 / n2
        c1 = - normal.dot(incident)
        # cos (incident_angle)
        c2sq = 1.0 - r * r * (1.0 - c1 * c1)
        # cos (refracted_angle) ** 2
        if c2sq.real < 0:
            # total internal reflection
            state = reflection(incident, normal, polarization_vector)
            return 0.0, state
            # no energy is abosrbed in the thinfilm
        c2 = sqrt(c2sq)
        # cos (refracted_angle)
        if c2.real > 1:
            # avoiding invalid solutions
            c2 = 1
        parallel_v, perpendicular_v, normal_parallel_plane = \
            parallel_orthogonal_components(polarization_vector, incident, normal)
        # parallel and perpendicular components of polarization vector
        # and orthogonal vector of the parallel plane
        ref_per = perpendicular_v.Length ** 2.0 / polarization_vector.Length ** 2.0
        # weight of perpendicular component: 0 < ref_per < 1
        if backside:
            # Ray intercepted on the backside of the transparent surface
            inc_angle = rad_to_deg(arccos(c2.real))
        else:
            inc_angle = rad_to_deg(arccos(c1))
        reflectance_matrix = properties['Matrix_reflectance_thin_film']
        r_matrix = reflectance_matrix(inc_angle, wavelength)
        # reflectance dependent of incidence angle and wavelength
        # We decide the polarization projection onto the parallel / perpendicular plane
        if myrandom() < ref_per:
            reflectance = calculate_reflectance(r_matrix, inc_angle, wavelength)[0]
            # reflectance for s-polarized (perpendicular) light
            perpendicular_polarized = True
            polarization_vector = normalize(perpendicular_v)
        else:
            reflectance = calculate_reflectance(r_matrix, inc_angle, wavelength)[1]
            # reflectance for p-polarized (parallel) light
            perpendicular_polarized = False
            polarization_vector = normalize(parallel_v)
        if myrandom() < reflectance:
            # ray reflected
            return 0.0, reflection(incident, normal, polarization_vector)
            # reflected_direction = simple_reflection(incident, normal).normalize()
            # if not perpendicular_polarized:
            #     # reflection changes the parallel component of incident polarization
            #     polarization_vector = simple_polarization_reflection(
            #         incident, normal, normal_parallel_plane, polarization_vector)
            # return 0.0, OpticalState(polarization_vector, reflected_direction,
            #                          Phenomenon.REFLEXION)  # TODO: Set solid
        else:
            # ray refracted: computing the refracted direction and energy absorbed in the thinfilm
            transmittance_matrix = properties['Matrix_transmittance_thin_film']
            t_matrix = transmittance_matrix(inc_angle, wavelength)
            # transmittance dependent of incidence angle and wavelength
            if perpendicular_polarized:
                transmittance = calculate_reflectance(t_matrix, inc_angle, wavelength)[0]
            else:
                transmittance = calculate_reflectance(t_matrix, inc_angle, wavelength)[1]
            factor_energy_absorbed = (1 - reflectance - transmittance) / (1 - reflectance)
            refracted_direction = incident * r.real + normal * (r.real * c1 - c2.real)
            refracted_direction.normalize()
            polarization_vector_out = outgoing_polarization(incident, polarization_vector,
                                                            refracted_direction, perpendicular_polarized)
            # if not perpendicular_polarized:
            #     # refraction changes the parallel component of incident polarization
            #     polarization_vector = \
            #         simple_polarization_refraction(
            #             incident, normal, normal_parallel_plane, c2, polarization_vector)
            return (factor_energy_absorbed,
                    OpticalState(polarization_vector_out, refracted_direction,
                                 Phenomenon.REFRACTION))  # TODO: Set solid

    def change_of_optical_state(self, ray, normal_vector):
        # the ray impacts on thin film material
        n1 = ray.current_medium().get_n(ray.wavelength)
        n_front = self.properties['index_of_refraction_front'](ray.wavelength)
        k_front = self.properties['extinction_coefficient_front'](ray.wavelength)
        n_back = self.properties['index_of_refraction_back'](ray.wavelength)
        k_back = self.properties['extinction_coefficient_back'](ray.wavelength)
        properties = self.properties
        if n1 == n_front + 1j * k_front:
            n2 = n_back + 1j * k_back
        else:
            n2 = n_front + 1j * k_front
        factor_energy_absorbed, optical_state = (
            self.calculate_state_thin_film(
                ray.current_direction(), normal_vector, n1, n2,
                ray.current_polarization(),
                properties, ray.wavelength))
        optical_state.extra_data['factor_energy_absorbed'] = \
            factor_energy_absorbed
        if optical_state.phenomenon == Phenomenon.REFRACTION:
            optical_state.material = self  # TODO: Set solid
        else:
            optical_state.material = ray.current_medium()  # TODO: Set solid
        return optical_state


vacuum_medium = SimpleVolumeMaterial("Vacuum", 1.0, 0.0)


@traced(logger)
class SurfaceMaterial(Material):
    """
    Subclass of Material for 2-dimensional materials (without volume)

    The `properties` parameter must be a dict with the physical properties
    describing the material. At least, the following must be provided:
    'probability_of_reflection': probability that a photon gets reflected,
    as a function of its wavelength.
    'probability_of_absorption': probability that a photon gets absorbed,
    as a function of its wavelength.
    'probability_of_transmittance': probability that a photon passes through
    the material, as a function of its wavelength.
    """

    def __init__(self, name, properties):
        super(SurfaceMaterial, self).__init__(name, properties)
        self.properties = properties

    @classmethod
    def create(cls, name, properties):
        _ = cls(name, properties)

    def compute_probabilities(self, ray):
        """
        Computes the tuple of probabilities that a ray hitting the surface gets reflected, absorbed or transmitted
        """
        properties = self.properties
        try:
            por = properties['probability_of_reflection'](ray.wavelength)
        except KeyError:
            por = 1.0
        try:
            poa = properties['probability_of_absorption'](ray.wavelength)
        except KeyError:
            poa = 1 - por
        try:
            pot = properties['probability_of_transmittance'](ray.wavelength)
        except KeyError:
            pot = 0.0

        return [por, poa, pot]

    def decide_phenomenon(self, ray):
        """
        Decides which phenomenon will take place when a ray hits the surface.
        """
        phenomena = [
            Phenomenon.REFLEXION,
            Phenomenon.ABSORPTION,
            Phenomenon.TRANSMITTANCE]

        probabilities = self.compute_probabilities(ray)
        phenomenon = np.random.choice(phenomena, 1, p=probabilities)[0]
        return phenomenon

    def change_of_optical_state(self, ray, normal_vector, nearby_material):
        phenomenon = self.decide_phenomenon(ray)
        properties = self.properties
        if phenomenon == Phenomenon.REFLEXION:
            polarization_vector = ray.current_polarization()
            incident = ray.current_direction()
            lambertian_weight, lambertian_kind = self.get_lambertian_weight_and_kind()
            state = reflection_hub(incident, normal_vector, polarization_vector,
                                   lambertian_weight, lambertian_kind)
            state.material = ray.current_medium()  # TODO: Set solid
            state.apply_dispersion(properties, normal_vector)
            return state
        if phenomenon == Phenomenon.ABSORPTION:
            if self.properties.get('thermal_material', False):
                return (OpticalState(Base.Vector(0.0, 0.0, 0.0),
                                     Base.Vector(0.0, 0.0, 0.0),
                                     Phenomenon.ENERGY_ABSORBED,
                                     self))  # TODO: Set solid
            else:
                return (OpticalState(Base.Vector(0.0, 0.0, 0.0),
                                     Base.Vector(0.0, 0.0, 0.0),
                                     Phenomenon.ABSORPTION,
                                     self))  # TODO: Set solid
        if phenomenon == Phenomenon.TRANSMITTANCE:
            # refraction in transparent layer
            n1 = ray.current_medium().get_n(ray.wavelength)
            n2 = nearby_material.get_n(ray.wavelength)
            if n1 == n2:  # transparent_simple_layer
                state = OpticalState(ray.current_polarization(),
                                     ray.current_direction(),
                                     Phenomenon.REFRACTION,
                                     nearby_material)  # TODO: Set solid
            else:
                lambertian_weight, lambertian_kind = self.get_lambertian_weight_and_kind()
                state = shure_refraction(ray.current_direction(), normal_vector,
                                         n1, n2, ray.current_polarization(),
                                         lambertian_weight, lambertian_kind)
                state.material = nearby_material  # TODO: Set solid
            return state

    @classmethod
    def from_plain_properties(cls, name, plain_properties):
        """
        Loads a material from its properties stored in a simple dictionary

        Parameters
        ----------
        name : str
        plain_properties : dict

        Returns
        -------
            Material
        """
        properties = Material.plain_properties_to_properties(plain_properties)
        material = cls(name, properties)
        return material

    def to_json(self):
        return json.dumps(
            {
                'name': self.name,
                'classname': self.__class__.__name__,
                'plain_properties': self.properties.get(
                    'plain_properties', None),
            }, cls=NumpyEncoder, indent=4
        )


@traced(logger)
class OpaqueSimpleLayer(SurfaceMaterial):
    """
    Subclass of `SurfaceMaterial` for completely opaque layers.
    """

    def __init__(self, name):
        plain_properties = {
            'probability_of_reflection': {
                'type': 'constant',
                'value': 0.0
            },
            'probability_of_absorption': {
                'type': 'constant',
                'value': 1.0
            },
            'probability_of_transmittance': {
                'type': 'constant',
                'value': 0.0
            },
            'thermal_material': {
                'type': 'scalar',
                'value': False
            }
        }
        properties = Material.plain_properties_to_properties(plain_properties)
        super(OpaqueSimpleLayer, self).__init__(name, properties)


@traced(logger)
class TransparentSimpleLayer(SurfaceMaterial):
    """
    Subclass of `SurfaceMaterial` for transparent layers.
    """

    def __init__(self, name, pot):
        plain_properties = {
            'probability_of_reflection': {
                'type': 'constant',
                'value': 1.0 - pot
            },
            'probability_of_absorption': {
                'type': 'constant',
                'value': 0.0
            },
            'probability_of_transmittance': {
                'type': 'constant',
                'value': pot
            },
            'thermal_material': {
                'type': 'scalar',
                'value': False
            },
            'specular_material': {
                'type': 'scalar',
                'value': True
            }
        }
        properties = Material.plain_properties_to_properties(plain_properties)
        super(TransparentSimpleLayer, self).__init__(name, properties)


@traced(logger)
class AbsorberSimpleLayer(SurfaceMaterial):
    """
    Subclass of `SurfaceMaterial` for absorber layers with behaviour independent of the wavelength.
    """

    def __init__(self, name, poa):
        plain_properties = {
            'probability_of_reflection': {
                'type': 'constant',
                'value': 1.0 - poa
            },
            'probability_of_absorption': {
                'type': 'constant',
                'value': poa
            },
            'probability_of_transmittance': {
                'type': 'constant',
                'value': 0.0
            },
            'thermal_material': {
                'type': 'scalar',
                'value': True
            },
            'specular_material': {
                'type': 'scalar',
                'value': True
            }
        }
        properties = Material.plain_properties_to_properties(plain_properties)
        super(AbsorberSimpleLayer, self).__init__(name, properties)


@traced(logger)
class AbsorberLambertianLayer(SurfaceMaterial):
    """
    Subclass of `SurfaceMaterial` for absorber layers with lambertian behaviour when reflecting rays.
    """

    def __init__(self, name, poa, weight=1, kind="Total"):
        plain_properties = {
            'probability_of_reflection': {
                'type': 'constant',
                'value': 1.0 - poa
            },
            'probability_of_absorption': {
                'type': 'constant',
                'value': poa
            },
            'probability_of_transmittance': {
                'type': 'constant',
                'value': 0.0
            },
            'thermal_material': {
                'type': 'scalar',
                'value': True
            },
            'lambertian_weight': {
                'type': 'scalar',
                'value': weight
            },
            'lambertian_kind': {
                'type': 'scalar',
                'value': kind
            },
        }
        properties = Material.plain_properties_to_properties(plain_properties)
        super(AbsorberLambertianLayer, self).__init__(name, properties)


@traced(logger)
class ReflectorSpecularLayer(SurfaceMaterial):
    """
    Subclass of `SurfaceMaterial` for reflector specular layers.
    """

    def __init__(self, name, por, sigma_1=None, sigma_2=None, k=None):
        plain_properties = {
            'probability_of_reflection': {
                'type': 'constant',
                'value': por
            },
            'probability_of_absorption': {
                'type': 'constant',
                'value': 1.0 - por
            },
            'probability_of_transmittance': {
                'type': 'constant',
                'value': 0.0
            },
            'thermal_material': {
                'type': 'scalar',
                'value': False
            },
            'specular_material': {
                'type': 'scalar',
                'value': True
            },
            'sigma_1': {
                'type': 'scalar',
                'value': sigma_1
            },
            'sigma_2': {
                'type': 'scalar',
                'value': sigma_2
            },
            'k': {
                'type': 'scalar',
                'value': k
            }
        }
        properties = Material.plain_properties_to_properties(plain_properties)
        super(ReflectorSpecularLayer, self).__init__(name, properties)


@traced(logger)
class ReflectorLambertianLayer(SurfaceMaterial):
    """
    Subclass of `SurfaceMaterial` for reflector layers with lambertian behaviour.
    """

    def __init__(self, name, por, weight=1, kind="Total"):
        plain_properties = {
            'probability_of_reflection': {
                'type': 'constant',
                'value': por
            },
            'probability_of_absorption': {
                'type': 'constant',
                'value': 1.0 - por
            },
            'probability_of_transmittance': {
                'type': 'constant',
                'value': 0.0
            },
            'thermal_material': {
                'type': 'scalar',
                'value': False
            },
            'lambertian_weight': {
                'type': 'scalar',
                'value': weight
            },
            'lambertian_kind': {
                'type': 'scalar',
                'value': kind
            },
        }
        properties = Material.plain_properties_to_properties(plain_properties)
        super(ReflectorLambertianLayer, self).__init__(name, properties)


@traced(logger)
class AbsorberTWModelLayer(SurfaceMaterial):
    """
    Subclass of `SurfaceMaterial` for materials using Tesfamichael-Wackelgard model.
    """

    def __init__(self, name, poa, b_constant, c_constant):
        plain_properties = {
            'probability_of_reflection': {
                'type': 'constant',
                'value': 1.0 - poa
            },
            'probability_of_absorption': {
                'type': 'constant',
                'value': poa
            },
            'probability_of_transmittance': {
                'type': 'constant',
                'value': 0.0
            },
            'thermal_material': {
                'type': 'scalar',
                'value': True
            },
            'TW_model': {
                'type': 'scalar',
                'value': True
            },
            'b_constant': {
                'type': 'scalar',
                'value': b_constant
            },
            'c_constant': {
                'type': 'scalar',
                'value': c_constant
            }
        }
        properties = Material.plain_properties_to_properties(plain_properties)
        super(AbsorberTWModelLayer, self).__init__(name, properties)

    @staticmethod
    def tw_absorptance_ratio(normal_vector, b_constant, c_constant, incident):
        """Angular Solar Absorptance model for selective absorber material.

        Given by the formula 1 - b * (1/cos - 1) ** c, based on:
        Tesfamichael, T., and Wackelgard, E., 2000, "Angular Solar Absorptance and
        Incident Angle Modifier of Selective Absorbers for Solar Thermal Collectors,"
        Sol. Energy, 68, pp. 335-341.

        Parameters
        ----------
        normal_vector : Base.Vector
            normal vector of the surface at the point of incidence
        b_constant : float
        c_constant : float
        incident : Base.Vector
            direction vector of the incident ray

        Returns
        -------
        float

        """
        # We assume the normal is normalized.
        normal = correct_normal(normal_vector, incident)
        c1 = - normal.dot(incident)
        inc_angle = rad_to_deg(arccos(c1))
        # incidence angle
        if inc_angle < 80.0:
            absorption_ratio = 1.0 - b_constant * abs((1.0 / c1 - 1.0)) ** c_constant
        else:
            y0 = 1.0 - b_constant * (1.0 / np.cos(80.0 * np.pi / 180.0) - 1.0) ** c_constant
            m = y0 / 10.0
            absorption_ratio = y0 - m * (inc_angle - 80.0)
        return absorption_ratio

    def change_of_optical_state(self, ray, normal_vector, nearby_material):
        properties = self.properties
        b_constant = properties['b_constant']
        c_constant = properties['c_constant']
        absorption_ratio = self.tw_absorptance_ratio(
            normal_vector, b_constant, c_constant, ray.current_direction())
        absorption = properties['probability_of_absorption'](ray.wavelength) * absorption_ratio
        reflectance = 1.0 - absorption
        if myrandom() < reflectance:
            polarization_vector = ray.current_polarization()
            state = reflection(ray.current_direction(), normal_vector, polarization_vector)
            state.material = ray.current_medium()  # TODO: Set solid
            return state
        else:
            # absorption in absorber material: the ray is killed
            return OpticalState(Base.Vector(0.0, 0.0, 0.0),
                                Base.Vector(0.0, 0.0, 0.0),
                                Phenomenon.ENERGY_ABSORBED, self)  # TODO: Set solid


@traced(logger)
class MetallicLayer(SurfaceMaterial):
    """
    Subclass of `SurfaceMaterial` for metallic layers.
    """


@traced(logger)
class MetallicSpecularLayer(MetallicLayer):
    """
    Subclass of `MetallicLayer` for metallic layers with specular properties.
    """

    def __init__(self, name, file_index_of_refraction,
                 sigma_1=None, sigma_2=None, k=None):
        # file_index_of_refraction with three columns: wavelenth in nm,
        # real(index of refraction), imaginary(index of refraction)
        data_refraction = load_from_txt_or_csv(file_index_of_refraction)
        # data_refraction = np.loadtxt(file_index_of_refraction, usecols=(0, 1, 2))
        wavelength_values = data_refraction[:, 0]
        n_values = data_refraction[:, 1]
        k_values = data_refraction[:, 2]
        plain_properties = {
            'index_of_refraction': {  # TODO: @Ramon Crec que aixo no s'empra per res
                'type': 'tabulated',
                'value': [wavelength_values, n_values]
            },
            'extinction_coefficient': {
                'type': 'tabulated',
                'value': [wavelength_values, k_values]
            },
            'sigma_1': {
                'type': 'scalar',
                'value': sigma_1
            },
            'sigma_2': {
                'type': 'scalar',
                'value': sigma_2
            },
            'k': {
                'type': 'scalar',
                'value': k
            }
        }
        properties = Material.plain_properties_to_properties(plain_properties)
        super(MetallicSpecularLayer, self).__init__(name, properties)

    def change_of_optical_state(self, ray, normal_vector, nearby_material):
        properties = self.properties
        polarization_vector = ray.current_polarization()
        n1 = ray.current_medium().get_n(ray.wavelength)
        n2 = self.get_n(ray.wavelength)
        incident = ray.current_direction()
        state = refraction(incident, normal_vector, n1, n2, polarization_vector)
        if state.phenomenon == Phenomenon.REFLEXION:
            state.material = ray.current_medium()  # TODO: Set solid
            state.apply_dispersion(properties, normal_vector)
            return state
        if state.phenomenon == Phenomenon.REFRACTION:
            # refraction in metallic layer: the ray is killed
            return OpticalState(Base.Vector(0.0, 0.0, 0.0),
                                Base.Vector(0.0, 0.0, 0.0),
                                Phenomenon.ABSORPTION, self)  # TODO: Set solid


@traced(logger)
class MetallicLambertianLayer(MetallicLayer):
    """
    Subclass of `MetallicLayer` for metallic layers with lambertian behaviour.
    """

    def __init__(self, name, file_index_of_refraction, weight=1, kind="Total"):
        # file_index_of_refraction with three columns:
        # wavelenth in nm, real(index of refraction),
        # imaginary(index of refraction)
        data_refraction = load_from_txt_or_csv(file_index_of_refraction)
        # data_refraction = np.loadtxt(file_index_of_refraction, usecols=(0, 1, 2))
        wavelength_values = data_refraction[:, 0]
        n_values = data_refraction[:, 1]
        k_values = data_refraction[:, 2]
        plain_properties = {
            'index_of_refraction': {
                'type': 'tabulated',
                'value': [wavelength_values, n_values]
            },
            'extinction_coefficient': {
                'type': 'tabulated',
                'value': [wavelength_values, k_values]
            },
            'lambertian_weight': {
                'type': 'scalar',
                'value': weight
            },
            'lambertian_kind': {
                'type': 'scalar',
                'value': kind
            },
        }
        properties = Material.plain_properties_to_properties(plain_properties)
        super(MetallicLambertianLayer, self).__init__(name, properties)

    def change_of_optical_state(self, ray, normal_vector, nearby_material):
        polarization_vector = ray.current_polarization()
        n1 = ray.current_medium().get_n(ray.wavelength)
        n2 = self.get_n(ray.wavelength)
        incident = ray.current_direction()
        lambertian_weight, lambertian_kind = self.get_lambertian_weight_and_kind()
        state = refraction(incident, normal_vector, n1, n2, polarization_vector,
                           lambertian_weight, lambertian_kind)
        if state.phenomenon == Phenomenon.REFLEXION:
            state.material = ray.current_medium()  # TODO: Set solid
            return state
        if state.phenomenon == Phenomenon.REFRACTION:
            # refraction in metallic layer: the ray is killed
            return OpticalState(Base.Vector(0.0, 0.0, 0.0),
                                Base.Vector(0.0, 0.0, 0.0),
                                Phenomenon.ABSORPTION, self)  # TODO: Set solid


@traced(logger)
class PolarizedCoatingLayer(SurfaceMaterial):
    """
    Subclass of `SurfaceMaterial` for polarized coating layers
    """

    def __init__(self, *args):
        super(PolarizedCoatingLayer, self).__init__(*args)

    def precompute_change_of_optical_state(self, ray, normal_vector):
        polarization_vector = ray.current_polarization()
        incident = ray.current_direction()
        wavelength = ray.wavelength
        properties = self.properties
        normal = correct_normal(normal_vector, incident)
        c1 = - normal.dot(incident)
        inc_angle = rad_to_deg(arccos(c1))
        # incidence angle
        parallel_v, perpendicular_v, normal_parallel_plane = \
            parallel_orthogonal_components(polarization_vector, incident, normal)
        ref_per = perpendicular_v.Length ** 2.0 / polarization_vector.Length ** 2.0
        reflectance_matrix = properties['Matrix_reflectance_coating']
        r_matrix = reflectance_matrix(inc_angle, wavelength)
        # reflectance dependent of incidence angle and wavelength
        # We decide the polarization projection onto the parallel / perpendicular plane
        if myrandom() < ref_per:
            reflectance = calculate_reflectance(r_matrix, inc_angle, wavelength)[0]
            # reflectance for s-polarized (perpendicular) light
            perpendicular_polarized = True
            polarization_vector = normalize(perpendicular_v)
        else:
            reflectance = calculate_reflectance(r_matrix, inc_angle, wavelength)[1]
            # reflectance for p-polarized (parallel) light
            perpendicular_polarized = False
            polarization_vector = normalize(parallel_v)
        if myrandom() < reflectance:
            # ray reflected
            state = reflection(incident, normal, polarization_vector)
            # reflected = simple_reflection(incident, normal).normalize()
            # if not perpendicular_polarized:
            #     # reflection changes the parallel component of incident polarization
            #     polarization_vector = simple_polarization_reflection(
            #         incident, normal, normal_parallel_plane, polarization_vector)
            # state = OpticalState(polarization_vector, reflected, Phenomenon.REFLEXION, self)
            state.material = ray.current_medium()  # TODO: Set solid
            state.apply_dispersion(properties, normal_vector)
            return state
        else:
            return (inc_angle, incident, perpendicular_polarized,
                    reflectance, normal_parallel_plane)


@traced(logger)
class PolarizedCoatingReflectorLayer(PolarizedCoatingLayer):
    """
    Subclass of `PolarizedCoatingLayer` for reflector layers.
    """

    def __init__(self, name, coating_file, sigma_1=None, sigma_2=None, k=None):
        # coating_material with four columns:
        # wavelenth in nm,
        # angle in deg.,
        # reflectance s-polarized (perpendicular),
        # reflectance p-polarized (parallel)
        # the values in coating_material should be in the corresponding
        # order columns
        data_material = load_from_txt_or_csv(coating_file)
        # data_material = np.loadtxt(coating_file, usecols=(0, 1, 2, 3))
        plain_properties = {
            'Matrix_reflectance_coating': {
                'type': 'matrix',
                'value': data_material
            },
            'sigma_1': {
                'type': 'scalar',
                'value': sigma_1
            },
            'sigma_2': {
                'type': 'scalar',
                'value': sigma_2
            },
            'k': {
                'type': 'scalar',
                'value': k
            }
        }
        properties = Material.plain_properties_to_properties(plain_properties)
        super(PolarizedCoatingReflectorLayer, self).__init__(name, properties)

    def change_of_optical_state(self, ray, normal_vector, nearby_material):
        new_state = self.precompute_change_of_optical_state(ray, normal_vector)
        if isinstance(new_state, OpticalState):
            return new_state
        else:
            # ray is killed in the coating reflector
            return OpticalState(Base.Vector(0.0, 0.0, 0.0),
                                Base.Vector(0.0, 0.0, 0.0),
                                Phenomenon.ABSORPTION)  # TODO: Set solid


@traced(logger)
class PolarizedCoatingAbsorberLayer(PolarizedCoatingLayer):
    """
    Subclass of `PolarizedCoatingLayer` for absorber layers.
    """

    def __init__(self, name, coating_file):
        # coating_material with four columns:
        # wavelenth in nm,
        # angle in deg.,
        # reflectance s-polarized (perpendicular),
        # reflectance p-polarized (parallel)
        # the values in coating_material should be in the corresponding order
        # columns
        data_material = load_from_txt_or_csv(coating_file)
        # data_material = np.loadtxt(coating_file, usecols=(0, 1, 2, 3))
        plain_properties = {
            'Matrix_reflectance_coating': {
                'type': 'matrix',
                'value': data_material
            },
            'probability_of_transmittance': {
                'type': 'constant',
                'value': 0
            },
        }
        properties = Material.plain_properties_to_properties(plain_properties)
        super(PolarizedCoatingAbsorberLayer, self).__init__(name, properties)

    def change_of_optical_state(self, ray, normal_vector, nearby_material):
        new_state = self.precompute_change_of_optical_state(ray, normal_vector)
        if isinstance(new_state, OpticalState):
            return new_state
        else:
            # ray refracted: ray energy will be absorbed in the coating absorber
            state = OpticalState(Base.Vector(0.0, 0.0, 0.0),
                                 Base.Vector(0.0, 0.0, 0.0),
                                 Phenomenon.ENERGY_ABSORBED, self)  # TODO: Set solid
            return state


@traced(logger)
class PolarizedCoatingTransparentLayer(PolarizedCoatingLayer):
    """
    Subclass of `PolarizedCoatingLayer` for transparent layers.
    """

    def __init__(self, name, coating_file):
        # coatingmaterial calculated by TMM method, six columns:
        # wavelength in nm,
        # angle in deg.,
        # reflectance s-polarized (perpendicular),
        # reflectance p-polarized (parallel),
        # transmittance s-polarized,
        # transmittance p-polarized
        # the values in coating_material should be in the corresponding
        # order columns
        ### data = np.loadtxt(coating_file)
        data = load_from_txt_or_csv(coating_file)
        data_reflectance = data[:, [0, 1, 2, 3]]
        data_transmittance = data[:, [0, 1, 4, 5]]
        plain_properties = {
            'Matrix_reflectance_coating': {
                'type': 'matrix',
                'value': data_reflectance
            },
            'Matrix_transmittance_coating': {
                'type': 'matrix',
                'value': data_transmittance
            },
        }
        properties = Material.plain_properties_to_properties(plain_properties)
        super(PolarizedCoatingTransparentLayer, self).__init__(name, properties)

    def change_of_optical_state(self, ray, normal_vector, nearby_material):
        polarization_vector = ray.current_polarization()
        incident = ray.current_direction()
        wavelength = ray.wavelength
        properties = self.properties
        normal = correct_normal(normal_vector, incident)
        n1 = ray.current_medium().get_n(ray.wavelength)
        n2 = nearby_material.get_n(ray.wavelength)
        r = n1 / n2
        c1 = - normal.dot(incident)
        # cos (incident_angle)
        c2sq = 1.0 - r * r * (1.0 - c1 * c1)
        # cos (refracted_angle) ** 2
        if c2sq.real < 0:
            # total internal reflection
            state = reflection(incident, normal, polarization_vector)
            state.material = ray.current_medium()  # TODO: Set solid
            return state
        c2 = sqrt(c2sq)
        if c2.real > 1:
            # avoiding invalid solutions
            c2 = 1
        new_state = self.precompute_change_of_optical_state(ray, normal_vector)
        if isinstance(new_state, OpticalState):
            return new_state
        else:
            (inc_angle, incident, perpendicular_polarized,
             reflectance, normal_parallel_plane) = new_state
            # ray refracted: computing the refracted direction and energy absorbed in coating
            transmittance_matrix = properties['Matrix_transmittance_coating']
            t_matrix = transmittance_matrix(inc_angle, wavelength)
            # transmittance dependent of incidence angle and wavelength
            if perpendicular_polarized:
                transmittance = calculate_reflectance(t_matrix, inc_angle, wavelength)[0]
            else:
                transmittance = calculate_reflectance(t_matrix, inc_angle, wavelength)[1]
            factor_energy_absorbed = (1 - reflectance - transmittance) / (1 - reflectance)
            refracted_direction = incident * r.real + normal * (r.real * c1 - c2.real)
            refracted_direction.normalize()
            polarization_vector_out = outgoing_polarization(incident, polarization_vector,
                                                            refracted_direction, perpendicular_polarized)
            # if not perpendicular_polarized:
            #     # refraction changes the parallel component of incident polarization
            #     polarization_vector = \
            #         simple_polarization_refraction(
            #             incident, normal, normal_parallel_plane, c2, polarization_vector)
            optical_state = OpticalState(polarization_vector_out, refracted_direction,
                                         Phenomenon.REFRACTION, nearby_material)  # TODO: Set solid
            optical_state.extra_data['factor_energy_absorbed'] = \
                factor_energy_absorbed
            return optical_state


@traced(logger)
class TwoLayerMaterial(Material):
    """
    Subclass of `Material` for surface materials formed by two layers (back and front)
    """

    def __init__(self, name, name_front_layer, name_back_layer):
        super(TwoLayerMaterial, self).__init__(name, {})
        self.name_front_layer = name_front_layer
        self.name_back_layer = name_back_layer
        self.front_material = Material.by_name[name_front_layer]
        self.back_material = Material.by_name[name_back_layer]

    def to_json(self):
        return json.dumps(
            {
                'name': self.name,
                'classname': 'TwoLayerMaterial',
                'name_front_layer': self.name_front_layer,
                'name_back_layer': self.name_back_layer
            }, cls=NumpyEncoder, indent=4
        )

    def change_of_optical_state(self, ray, normal_vector, nearby_material):
        if ray.current_direction().dot(normal_vector) < 0:
            # Ray intercepted on the frontside of the surface
            material = self.front_material
        else:
            # Ray intercepted on the backside of the surface
            material = self.back_material
        return material.change_of_optical_state(ray, normal_vector, nearby_material)



"""Module otsun.ray for the modelization of light rays

The module defines the class `Ray`
"""

from autologging import traced
import logging
logger = logging.getLogger(__name__)
from .materials import vacuum_medium, PVMaterial, SurfaceMaterial, TwoLayerMaterial, PolarizedThinFilm
from .optics import Phenomenon, OpticalState
import numpy as np
import Part
from FreeCAD import Base

try:
    part_line = Part.LineSegment
except AttributeError:
    part_line = Part.Line

# Zero energy level
LOW_ENERGY = 1E-6


def _center(bb):
    return Base.Vector((bb.XMin+bb.XMax)/2, (bb.YMin+bb.YMax)/2, (bb.ZMin+bb.ZMax)/2)


def _interval_intersects(x1, x2, y1, y2):
    return x1 <= y2 and y1 <= x2


def _bb_intersects(bb1, bb2):
    return (_interval_intersects(bb1.XMin, bb1.XMax, bb2.XMin, bb2.XMax) and
            _interval_intersects(bb1.YMin, bb1.YMax, bb2.YMin, bb2.YMax) and
            _interval_intersects(bb1.ZMin, bb1.ZMax, bb2.ZMin, bb2.ZMax))


def _distance_point_to_line(point, line_point, line_vector):
    v = point-line_point
    cross = v.cross(line_vector)
    return cross.Length


def _distance_point_to_ray(point, ray_origin, ray_vector):
    v = point - ray_origin
    if v.dot(ray_vector) >= 0:
        return _distance_point_to_line(point, ray_origin, ray_vector)
    return v.Length


def _line_may_intersect_bb(bb, line_point, line_vector):
    center = _center(bb)
    d = _distance_point_to_line(center, line_point, line_vector)
    return d <= bb.DiagonalLength/2


def _ray_may_intersect_bb(bb, ray_origin, ray_vector):
    center = _center(bb)
    d = _distance_point_to_ray(center, ray_origin, ray_vector)
    return d <= bb.DiagonalLength/2


@traced(logger)
class Ray(object):
    """
    Class used to model a sun ray

    It keeps information of the path it describes and its polarization as it interacts
    with the scene and its energy.

    Parameters
    ----------
    scene : Scene
        Scene where the ray moves
    origin : Base.Vector
        Point of emission of the ray
    direction : Base.Vector
        Direction of the ray when it was emitted
    wavelength : float
        Wavelength of the ray
    energy : float
        Initial energy of the ray
    polarization_vector : Base.Vector
        Initial polarization vector of the ray

    Attributes
    ----------
    points : list of Base.Vector
        Points where the ray passes at each iteration
    optical_states : list of OpticalState
        OpticalStates of the rays at each iteration
    last_normal : Base.Vector
        Last vector normal to the surface where the ray hits (used for PV)
    wavelength : float
        Wavelength of ray
    energy : float
        Current energy of ray
    polarization_vectors : list of Base.Vector
        Polarization of the ray at each iteration
    finished : bool
    Th_absorbed : bool
    PV_values : list of tuple of floats
        List where each entry is a PV_value (a tuple of floats)
    PV_absorbed : list of float
        List of values of absorbed PV energy
    """

    def __init__(self, scene, origin, direction,
                 wavelength, energy, polarization_vector):
        self.scene = scene
        self.points = [origin]
        state = OpticalState(
            polarization_vector,
            direction,
            Phenomenon.JUST_STARTED,
            vacuum_medium
        )
        self.optical_states = [state]
        self.current_solid = None
        self.last_normal = None
        self.last_touched_face = None
        self.wavelength = wavelength
        self.energy = energy
        self.polarization_vectors = [polarization_vector]
        self.finished = False
        self.Th_absorbed = False
        self.PV_values = []
        self.PV_absorbed = []

    def __str__(self):
        return "Pos.: %s, OS: %s, Energy: %s" % (
            self.points[-1], self.optical_states[-1], self.energy
        )

    def current_medium(self):
        """
        Get medium where ray is currently traveling

        Returns
        -------
            otsun.VolumeMaterial
        """
        return self.optical_states[-1].material

    def current_direction(self):
        """
        Get current direction

        Returns
        -------
            Base.Vector
        """
        return self.optical_states[-1].direction

    def current_polarization(self):
        """
        Get current polarization

        Returns
        -------
            Base.Vector
        """
        return self.optical_states[-1].polarization

    def weighted_distance(self, p0, pair):
        if id(pair[1]) in self.scene.materials:
            return p0.distanceToPoint(pair[0])
        else:
            return p0.distanceToPoint(pair[0]) + self.scene.epsilon

    def next_intersection(self):
        """
        Finds next intersection of the ray with the scene

        Returns a pair (point,face), where point is the point of intersection
        and face is the face that intersects.
        If no intersection is found, it returns a "point at infinity" and None.

        Returns
        -------
        point : Base.Vector
            Point of intersection
        face : Part.Face or None
            Face where it intersects
        """
        max_distance = 5 * self.scene.diameter
        p0 = self.points[-1]
        direction = self.current_direction()
        # p0 = p0 + direction * self.scene.epsilon
        intersections = []
        p1 = p0 + direction * max_distance
        segment = part_line(p0, p1)
        shape_segment = segment.toShape()
        bb1 = shape_segment.BoundBox
        candidates = self.scene.faces
        feasible_faces = 0
        filtered_faces_1 = 0
        filtered_faces_2 = 0
        feasible_but_empty = 0
        for face in candidates:
            bb2 = face.BoundBox
            if not _ray_may_intersect_bb(bb2, p0, direction):
                filtered_faces_1 += 1
                continue
            if not _bb_intersects(bb1, bb2):
                filtered_faces_2 += 1
                continue
            feasible_faces += 1
            punts = face.section(shape_segment).Vertexes
            # if face == self.last_touched_face:
            #     punts = [punt for punt in punts if p0.distanceToPoint(punt.Point) > self.scene.epsilon]
            if not punts:
                feasible_but_empty += 1
                continue
                # logger.debug("feasible face but empty intersection")
            # material = self.scene.materials[face]
            if (id(face) in self.scene.materials) & (face != self.last_touched_face):
                for punt in punts:
                    intersections.append([punt.Point, face])
            else:
                for punt in punts:
                    if p0.distanceToPoint(punt.Point) > self.scene.epsilon:
                        intersections.append([punt.Point, face])

        logger.debug(f"Found {len(intersections)} points in {feasible_faces} faces. Filtered {filtered_faces_1}+{filtered_faces_2}. Feasible but empty {feasible_but_empty}")
#        intersections = [punt_cara for punt_cara in intersections if
#                         p0.distanceToPoint(punt_cara[0]) > self.scene.epsilon]
        if not intersections:
            logger.debug("No true intersection found with scene")
            return p1, None
        closest_pair = min(intersections,
                           key=lambda pair: self.weighted_distance(p0, pair))
        return tuple(closest_pair)

    def next_state_solid_and_normal(self, face):
        """
        Computes the next optical state after the ray hits the face, and the normal vector

        Parameters
        ----------
        face

        Returns
        -------
        state : OpticalState
            Optical State that the ray will have after hitting the face
        next_solid : Part.Solid
            Solid where the ray will travel after hitting the face
        normal : Base.Vector
            Normal vector to the face where the ray hits
        """
        current_direction = self.current_direction()
        current_point = self.points[-1]
        uv = face.Surface.parameter(current_point)
        normal = face.normalAt(uv[0], uv[1])
        normal.normalize()
        nearby_solid = self.scene.next_solid_at_point_in_direction(current_point, normal, current_direction)
        nearby_material = self.scene.materials.get(id(nearby_solid), vacuum_medium)
        if id(face) in self.scene.materials:
            # face is active
            material = self.scene.materials[id(face)]
            # point_plus_delta = current_point + current_direction * 2 * self.scene.epsilon
            state = material.change_of_optical_state(self, normal, nearby_material)
            # TODO polarization_vector
        else:
            # face is not active
            # point_plus_delta = current_point + current_direction * 2 * self.scene.epsilon
            # next_solid = self.scene.solid_at_point(point_plus_delta)
            # nearby_material = self.scene.materials.get(next_solid, vacuum_medium)
            if isinstance(nearby_material, (SurfaceMaterial, TwoLayerMaterial)):
                # solid with surface material on all faces
                state = nearby_material.change_of_optical_state(self, normal, nearby_material)
            else:
                state = nearby_material.change_of_optical_state(self, normal)
            # TODO polarization_vector
        logger.debug(state)
        next_solid = None
        if state.phenomenon in [Phenomenon.ENERGY_ABSORBED, Phenomenon.ABSORPTION]:
            pass
        elif state.phenomenon == Phenomenon.REFLEXION:
            next_solid = self.current_solid
        elif state.phenomenon in [Phenomenon.REFRACTION, Phenomenon.TRANSMITTANCE]:
            next_solid = nearby_solid
        return state, next_solid, normal

    def update_energy(self):
        material = self.current_medium()
        if (material.name == "Vacuum" and
                'vacuum_material' in self.scene.extra_data and
                len(self.points) > 2
        ):
            # implementation of dispersive vacuum after first interaction with scene
            material = self.scene.extra_data['vacuum_material']
        # TODO: @Ramon
        point_1 = self.points[-1]
        point_2 = self.points[-2]
        if material.properties.get('extinction_coefficient', None):
            alpha = material.properties['extinction_coefficient'](self.wavelength) * \
                    4 * np.pi / (self.wavelength / 1E6)  # mm-1
            if alpha:
                d = point_1.distanceToPoint(point_2)
                self.energy = self.energy * np.exp(- alpha * d)
        if material.properties.get('attenuation_coefficient', None):
            alpha = material.properties['attenuation_coefficient'](self.wavelength)  # mm-1
            if alpha:
                d = point_1.distanceToPoint(point_2)
                self.energy = self.energy * np.exp(- alpha * d)

    def run(self, max_hops=200):
        """
        Makes the ray propagate

        Makes the sun ray propagate until it gets absorbed, it exits the scene,
        or gets caught in multiple (> max_hops) reflections.

        Parameters
        ----------
        max_hops : int
            Maximum number of iterations
        """
        count = 0
        while (not self.finished) and (count < max_hops):
            logger.debug("Ray running. Hop %s, %s, Solid %s", count, self,
                        self.scene.name_of_solid.get(self.current_solid, "Void"))
            count += 1

            # Find next intersection
            # Update points
            point, face = self.next_intersection()
            self.last_touched_face = face
            self.points.append(point)
            if not face:
                self.finished = True
                self.Th_absorbed = False
                break

            # Update energy
            energy_before = self.energy
            self.update_energy()

            # Treat PV material
            current_material = self.current_medium()
            if isinstance(current_material, PVMaterial):
                PV_absorbed_energy, PV_value = current_material.get_PV_data(self, energy_before)
                self.PV_values.append(PV_value)
                self.PV_absorbed.append(PV_absorbed_energy)

            # Update optical state
            state, next_solid, normal = self.next_state_solid_and_normal(face)

            # TODO: The following line is not elegant.
            #  Provisional solution for updating energy when passed a thin film
            if 'factor_energy_absorbed' in state.extra_data:
                factor = state.extra_data['factor_energy_absorbed']
                self.energy = self.energy * (1 - factor)

            # Update optical_states
            self.optical_states.append(state)
            self.last_normal = normal
            self.current_solid = next_solid

            # Test for finish
            if state.phenomenon == Phenomenon.ABSORPTION:
                self.finished = True
            if state.phenomenon == Phenomenon.ENERGY_ABSORBED:
                self.Th_absorbed = True
                self.finished = True
            if self.energy < LOW_ENERGY:
                self.finished = True
        logger.debug("Ray stopped. Hop %s, %s, Solid %s", count, self,
                    self.scene.name_of_solid.get(self.current_solid, "Void"))

    def add_to_document(self, doc):
        """
        Draws the ray in a FreeCAD document

        Parameters
        ----------
        doc : App.Document
        """
        lshape_wire = Part.makePolygon(self.points)
        my_shape_ray = doc.addObject("Part::Feature", "Ray")
        my_shape_ray.Shape = lshape_wire



try:
    import freecad
    # Some distribuitions of FreeCAD need this import to set up the paths
except ImportError:
    pass
from .math import *
from .optics import *
from .outputs import *
from .materials import *
from .ray import *
from .source import *
from .scene import *
from .experiments import *
from .movements import *

import importlib.metadata
from pathlib import Path
import toml

import logging
logger = logging.getLogger(__name__)

try:
    __version__ = importlib.metadata.version('otsun')
except importlib.metadata.PackageNotFoundError:
    __version__ = "unknown"
    # adopt path to your pyproject.toml
    pyproject_toml_file = Path(__file__).parent.parent.parent / "pyproject.toml"
    print(pyproject_toml_file)
    if pyproject_toml_file.exists() and pyproject_toml_file.is_file():
        data = toml.load(pyproject_toml_file)
        # check project.version
        if "project" in data and "version" in data["project"]:
            __version__ = data["project"]["version"]
        # check tool.poetry.version
        elif "tool" in data and "poetry" in data["tool"] and "version" in data["tool"]["poetry"]:
            __version__ = data["tool"]["poetry"]["version"]



"""Module otsun.movements for computing movements

The module defines the classes `Joint` (with subclasses `AxialJoint` and `CentralJoint`) and `MultiTracking`
"""
from FreeCAD import Base
from .math import one_orthogonal_vector, projection_on_orthogonal_of_vector, EPSILON
from numpy import pi
from numpy.random import normal as random_normal

from autologging import traced
import logging
logger = logging.getLogger(__name__)


def orientation(u, v, w):
    """Computes the orientation (+/-1) of a basis"""
    det = u.dot(v.cross(w))
    if det >= 0:
        return +1
    else:
        return -1


def signed_angle(axis, v1, v2):
    """
    Computes the (signed) angle from v1 to v2, wrt the axis vector
    """
    angle = v1.getAngle(v2)
    angle *= orientation(axis, v1, v2)
    return angle


def apply_movement_to_vector(movement, vector):
    pA = Base.Vector(0, 0, 0)
    pB = pA + vector
    pAprime = movement.multVec(pA)
    pBprime = movement.multVec(pB)
    return pBprime - pAprime


def axial_rotation_from_axis_and_angle(axis_origin, axis_dir, angle):
    """
    Returns a rotation with given axis and angle
    """
    z_axis = Base.Vector(0, 0, 1)
    local_cs = Base.Placement(axis_origin, Base.Rotation(z_axis, axis_dir))
    return local_cs.multiply(
        Base.Placement(Base.Vector(), Base.Rotation(angle * 180.0 / pi, 0, 0)).multiply(local_cs.inverse()))


def axial_rotation_from_vector_and_image(origin, vector0, vector1, error=0):
    """
    Returns a rotation that transforms the ray with origin and vector equals to vector0 to the ray with same
    origin and vector vector1.
    """
    try:
        normal = vector0.cross(vector1)
        normal.normalize()
    except Base.FreeCADError:
        # we assume now that the two vectors are collinear
        normal = one_orthogonal_vector(vector0)
    angle = vector0.getAngle(vector1)
    angle += error
    return axial_rotation_from_axis_and_angle(origin, normal, angle)

def biaxial_rotation_from_vector_and_image(origin, vector_v, vector_h, current_normal, desired_normal,
                                           error_v=0, error_h=0):
    # First rotate around vector_h, so that the image of vector_v is orthogonal
    # to desired_normal
    image_vector_horizontal = desired_normal.cross(vector_h)
    g1 = axial_rotation_from_vector_and_image(origin, vector_v, image_vector_horizontal, error_v)
    new_normal = apply_movement_to_vector(g1, current_normal)
    g2 = axial_rotation_from_vector_and_image(origin, new_normal, desired_normal, error_h)
    return g2.multiply(g1)


def get_labels(obj):
    """Gets the different labels of a given FreeCAD object"""
    label = obj.Label
    start = label.find("(")
    end = label.find(")")
    if start < 0 or end < 0:
        return []
    string = label[start + 1:end]
    return string.split(',')


@traced(logger)
class Joint:
    """
    Base class to represent joints
    """
    pass


@traced(logger)
class AxialJoint(Joint):
    """
    Class used to represent joints that rotate around an axis
    """

    def __init__(self, axis_origin, axis_vector, sigma=None):
        self.axis_origin = axis_origin
        self.axis_vector = axis_vector
        self.sigma = sigma
        if self.sigma is not None:
            self.error = random_normal(scale=self.sigma/1000)
        else:
            self.error = 0

    def compute_rotation_to_point(self, target, normal, light_direction):
        """
        Computes a rotation so that ray in the direction of light_direction is reflected and hits the target.
        """
        pointing = projection_on_orthogonal_of_vector(target - self.axis_origin, self.axis_vector)
        pointing.normalize()
        light_unit = projection_on_orthogonal_of_vector(light_direction, self.axis_vector)
        if light_unit.Length > EPSILON:
            light_unit.normalize()
            desired_normal = pointing - light_unit
            angle = signed_angle(self.axis_vector, normal, desired_normal)  # normal.getAngle(desired_normal)
        else:
            angle = 0
        angle += self.error
        return axial_rotation_from_axis_and_angle(self.axis_origin, self.axis_vector, angle)

    def compute_rotation_to_direction(self, normal, light_direction):
        """
        Computes a rotation so that ray in the direction of light_direction is returned to its origin.
        """
        light_unit = projection_on_orthogonal_of_vector(light_direction, self.axis_vector)
        light_unit.normalize()
        desired_normal = light_unit * (-1.0)
        angle = signed_angle(self.axis_vector, normal, desired_normal)  # normal.getAngle(desired_normal)
        angle += self.error
        return axial_rotation_from_axis_and_angle(self.axis_origin, self.axis_vector, angle)


@traced(logger)
class TwoOrthogonalAxialJoint(Joint):
    """
    Class used to represent joints with two orthogonal axis
    """

    def __init__(self, axis_origin, axis_vector_vertical, axis_vector_horizontal, sigma_v=None, sigma_h=None):
        self.axis_origin = axis_origin
        self.axis_vector_vertical = axis_vector_vertical
        self.axis_vector_horizontal = axis_vector_horizontal
        self.sigma_v = sigma_v
        self.sigma_h = sigma_h
        if self.sigma_v is not None:
            self.error_v = random_normal(scale=self.sigma_v/1000)
        else:
            self.error_v = 0
        if self.sigma_h is not None:
            self.error_h = random_normal(scale=self.sigma_h/1000)
        else:
            self.error_h = 0


    def compute_rotation_to_point(self, target, normal, light_direction):
        """
        Computes a rotation so that ray in the direction of light_direction is reflected and hits the target.
        """
        pointing = target - self.axis_origin
        pointing.normalize()
        light_unit = light_direction * 1.0
        light_unit.normalize()
        desired_normal = pointing - light_unit
        return biaxial_rotation_from_vector_and_image(self.axis_origin, self.axis_vector_vertical,
                                                      self.axis_vector_horizontal, normal, desired_normal,
                                                      self.error_v, self.error_h)

    def compute_rotation_to_direction(self, normal, light_direction):
        """
        Computes a rotation so that ray in the direction of light_direction is returned to its origin.
        """
        light_unit = light_direction * 1.0
        light_unit.normalize()
        desired_normal = light_unit * (-1.0)
        return biaxial_rotation_from_vector_and_image(self.axis_origin, self.axis_vector_vertical,
                                                      self.axis_vector_horizontal, normal, desired_normal,
                                                      self.error_v, self.error_h)


@traced(logger)
class CentralJoint(Joint):
    """
    Class used to represent joints that rotate around a point
    """

    def __init__(self, center):
        self.center = center

    def compute_rotation_to_point(self, target, normal, light_direction):
        """
        Computes a rotation so that ray in the direction of light_direction is reflected and hits the target.
        """
        pointing = target - self.center
        pointing.normalize()
        light_unit = light_direction * 1.0
        light_unit.normalize()
        desired_normal = pointing - light_unit
        return axial_rotation_from_vector_and_image(self.center, normal, desired_normal)

    def compute_rotation_to_direction(self, normal, light_direction):
        """
        Computes a rotation so that ray in the direction of light_direction is returned to its origin.
        """
        light_unit = light_direction * 1.0
        light_unit.normalize()
        desired_normal = light_unit * (-1.0)
        return axial_rotation_from_vector_and_image(self.center, normal, desired_normal)


@traced(logger)
class MultiTracking:
    """
    Class that represents how objects are linked to joints and implements their movement.
    """

    def __init__(self, source_direction, scene, analyze_labels=True):
        self.scene = scene
        self.source_direction = source_direction
        self.object_joint_map = {}
        self.object_normal_map = {}
        self.object_target_map = {}
        self.object_movements_map = {}
        if analyze_labels:
            self.get_data_from_objects()
            self.compute_movements()

    def get_object_from_label(self, label):
        return [obj2 for obj2 in self.scene.objects if obj2.Label == label][0]

    @staticmethod
    def get_dimension_of_object(obj):
        obj_type = obj.Shape.ShapeType
        if obj_type == 'Vertex':
            return 0
        if obj_type in ['Wire', 'Edge']:
            return 1
        return None

    @staticmethod
    def get_point_from_vertex(vertex):
        return vertex.Shape.Vertexes[0].Point

    @staticmethod
    def get_point_and_vector_from_edge(edge):
        start = edge.Shape.Vertexes[0].Point
        end = edge.Shape.Vertexes[1].Point
        return start, end - start

    @staticmethod
    def get_center_two_edges(ob1, ob2):
        sh1 = ob1.Shape
        sh2 = ob2.Shape
        d, pairs_data, _ = sh1.distToShape(sh2)
        pairs = pairs_data[0]
        return (pairs[0]+pairs[1])/2

    def get_target_from_label(self, label):
        """
        Finds the target of an object according to its label
        """
        target_obj = [obj2 for obj2 in self.scene.objects if obj2.Label == label][0]
        if target_obj.Shape.ShapeType == "Vertex":
            target = target_obj.Shape.Vertexes[0].Point
            return target

    def get_deviation(self, axis_label):
        try:
            return self.scene.extra_data['axis_deviation'][axis_label]
        except KeyError:
            return None

    def get_data_from_objects(self):
        """
        Gets all data of joints from the objects in the scene
        """
        cached_joints = dict()
        for obj in self.scene.objects:
            labels = get_labels(obj)
            if len(labels) < 2:
                # no movement
                continue
            objects = [self.get_object_from_label(label) for label in labels[1:]]
            dimensions = [self.get_dimension_of_object(ob) for ob in objects]
            if dimensions in [[1, 1, 1, 0], [1, 1, 1]]:  # Biaxial joint
                _, axis1_vector = self.get_point_and_vector_from_edge(objects[0])
                _, axis2_vector = self.get_point_and_vector_from_edge(objects[1])
                center = self.get_center_two_edges(objects[0], objects[1])
                _, normal = self.get_point_and_vector_from_edge(objects[2])
                sigma1 = self.get_deviation(objects[0].Label)
                sigma2 = self.get_deviation(objects[1].Label)
                if (labels[1], labels[2]) not in cached_joints:
                    cached_joints[(labels[1], labels[2])] = TwoOrthogonalAxialJoint(
                        center, axis2_vector, axis1_vector, sigma2, sigma1)
                self.object_joint_map[obj] = cached_joints[(labels[1], labels[2])]
                self.object_normal_map[obj] = normal
                pass
            elif dimensions in [[1, 1, 0], [1, 1]]:  # Axial joint
                axis_point, axis_vector = self.get_point_and_vector_from_edge(objects[0])
                _, normal = self.get_point_and_vector_from_edge(objects[1])
                sigma = self.get_deviation(objects[0].Label)
                if labels[1] not in cached_joints:
                    cached_joints[labels[1]] = AxialJoint(axis_point, axis_vector, sigma)
                self.object_joint_map[obj] = cached_joints[labels[1]]
                self.object_normal_map[obj] = normal
                pass
            elif dimensions in [[0, 1, 0], [0, 1]]:  # Central joint
                center = self.get_point_from_vertex(objects[0])
                _, normal = self.get_point_and_vector_from_edge(objects[1])
                self.object_joint_map[obj] = CentralJoint(center)
                self.object_normal_map[obj] = normal
            else:
                raise Exception("Malformed joint characterization")
            if dimensions[-1] == 0:
                self.object_target_map[obj] = self.get_point_from_vertex(objects[-1])
            else:
                self.object_target_map[obj] = None

    def compute_movements(self):
        """
        Computes the movements so that the objects point in the right direction according to their target
        """
        for obj in self.object_joint_map:
            joint = self.object_joint_map[obj]
            normal = self.object_normal_map[obj]
            target = self.object_target_map[obj]
            if target is None:
                # source tracking
                movement = joint.compute_rotation_to_direction(normal, self.source_direction)
            else:
                movement = joint.compute_rotation_to_point(target, normal, self.source_direction)
            self.object_movements_map[obj] = movement

    def make_movements(self):
        """
        Makes the computed movements
        """
        for obj in self.object_movements_map:
            movement = self.object_movements_map[obj]
            obj.Placement = movement.multiply(obj.Placement)
            for element, el_obj in self.scene.element_object_dict.items():
                if obj == el_obj:
                    element.Placement = movement.multiply(element.Placement)
        self.scene.recompute_boundbox()

    def undo_movements(self):
        """
        Undo the movements so that they are in their original position
        """
        for obj in self.object_movements_map:
            movement = self.object_movements_map[obj]
            movement = movement.inverse()
            obj.Placement = movement.multiply(obj.Placement)
            for element, el_obj in self.scene.element_object_dict.items():
                if obj == el_obj:
                    element.Placement = movement.multiply(element.Placement)
        self.scene.recompute_boundbox()



"""
Module otsun.math with mathematical helper functions
"""

import numpy as np
from FreeCAD import Base
import random
import time
from functools import wraps

EPSILON = 1E-6
# Tolerance for considering equal to zero
INF = 1E20
# Infinite

try:
    from functools import lru_cache
except ImportError:
    from backports.functools_lru_cache import lru_cache


def polar_to_cartesian(phi, theta):
    """Convert polar coordinates of unit vector to cartesian

    Parameters
    ----------
    phi : float
        phi angle (ISO 31-11) in degrees
    theta : float
        theta angle (ISO 31-11) in degrees

    Returns
    -------
    Base.Vector

    """
    rad = np.pi / 180.0
    x = np.sin(theta * rad) * np.cos(phi * rad)
    y = np.sin(theta * rad) * np.sin(phi * rad)
    z = np.cos(theta * rad)
    return Base.Vector(x, y, z)


def rad_to_deg(angle):
    """Converts radians to degrees"""
    return angle * 180.0 / np.pi


# ---
# Helper functions for input of functions
# ---


def constant_function(c):
    """Create a constant function

    Parameters
    ----------
    c : float
        constant to return
    Returns
    -------
    function
        Constant function equal to `c`
    """
    return lambda x: c


def tabulated_function(xvalues, yvalues):
    """Create a linear interpolating function from tabulated values

    Parameters
    ----------
    xvalues : list of float
        x coordinates of the tabulated values
    yvalues : list of float
        y coordinates of the tabulated values

    Returns
    -------
    function
        Function that interpolates by straight line segments the input data
    """

    # @memoize
    @lru_cache(maxsize=None)
    def this_tabulated_function(x):
        return np.interp(x, xvalues, yvalues)

    return this_tabulated_function


# # ---
# # Helper function for random Linear congruential generator
# # ---
# _previous = None
# def random_congruential(seed=None):
#     """Random Linear congruential generator based on  MTH$RANDOM
#
#     Parameters
#     ----------
#     seed : float
#         seed to use in the generation of random numbers
#
#     Returns
#     -------
#     float
#
#     """
#     # http://daviddeley.com/random/random4.htm
#     a = 69069.0
#     c = 1.0
#     m = 2.0 ** 32.0
#     rm = 1.0 / m
#     global _previous
#     if not seed:
#         if not _previous:
#             _previous = time.time()
#     else:
#         _previous = seed
#     _previous = np.remainder(_previous * a + c, m)
#     return _previous * rm


# ---
# Define the random algorithm
# ---
# myrandom = random_congruential
myrandom = random.random

# ---
# Helper function for Cumulative Function Distribution and Randomly generatting distribution
# ---


def cdf_from_pdf_file(data_file):
    """
    Computes CDF from PDF values stored in a file

    Creates a Cumulative Distribution Function from Probability Density
    Function data file. Each line must be a pair of numbers x y=pdf(x).
    It returns the CDF as two lists; first on is the list of x-values,
    second one is the list of corresponding CDF values.

    Parameters
    ----------

    data_file: file or str
        file or filename where PDF values are stored

    Returns
    -------
    list of float, list of float
        x-values and y-values of CDF
    """
    data_array = np.loadtxt(data_file, usecols=(0, 1))
    x = data_array[:, 0]
    y = data_array[:, 1]
    x_cdf = x
    n = np.size(y)
    y_ii = []
    for i in np.arange(n - 1):
        y_i = (y[i + 1] + y[i]) / 2.0 * (x[i + 1] - x[i])
        y_ii = np.append(y_ii, y_i)
    y_ii = np.append(y_ii, y_ii[-1])
    k_integration = np.trapz(y_ii, x_cdf)
    y_cdf = np.cumsum(y_ii) / k_integration
    return x_cdf, y_cdf / y_cdf[-1]


def pick_random_from_cdf(cdf):
    """
    Pick a random value according to a given CDF.

    We apply the Inverse transform sampling: https://en.wikipedia.org/wiki/Inverse_transform_sampling

    Parameters
    ----------
    cdf : tuple of list of float
        First list is list of x-values; second one is list of values of CDF

    Returns
    -------
    float
    """
    return np.interp(random.random(), cdf[1], cdf[0])


def parallel_orthogonal_components(vector, incident, normal):
    """Decomposition of vector in components

    Given `vector` (a polarization),
    `incident` (direction of a ray) and
    `normal` (vector orthogonal to a plane),
    decompose `vector` it in
    a component contained in the reflection (parallel) plane (det. by normal and incident): p-polarized (parallel) light
    a component contained in the orthogonal plane to the reflection plane: s-polarized (perpendicular) light
    also returns the normal vector to the reflection plane

    Parameters
    ----------
    vector : Base.Vector
    incident : Base.Vector
    normal : Base.Vector

    Returns
    -------
    parallel : Base.Vector
    orthogonal : Base.Vector
    normal_of_parallel_plane: Base.Vector
    """
    polarization_vector = vector
    normal_parallel_plane = incident.cross(normal)
    # orthogonal vector to reflection plane (parallel_plane)
    if normal_parallel_plane.Length < EPSILON:
        normal_parallel_plane = one_orthogonal_vector(normal)
    normal_parallel_plane.normalize()
    normal_perpendicular_plane = incident.cross(normal_parallel_plane)
    # orthogonal vector to perpendicular_plane
    parallel_v = (polarization_vector -
                  normal_parallel_plane * polarization_vector.dot(normal_parallel_plane))
    # parallel_v is the projection of polarization_vector onto parallel_plane
    perpendicular_v = (polarization_vector -
                       normal_perpendicular_plane * polarization_vector.dot(normal_perpendicular_plane))
    # perpendicular_v is the projection of polarization_vector onto the perpendicular_plane
    return parallel_v, perpendicular_v, normal_parallel_plane


def two_orthogonal_vectors(vector):
    """Gives two orthogonal vectors of a vector

    Given `vector` find two orthogonal vectors

    Parameters
    ----------
    vector : Base.Vector

    Returns
    -------
    orthogonal_1 : Base.Vector
    orthogonal_2 : Base.Vector
    """
    orthogonal_1 = one_orthogonal_vector(vector)
    orthogonal_2 = vector.cross(orthogonal_1)
    return orthogonal_1.normalize(), orthogonal_2.normalize()


def one_orthogonal_vector(vector):
    """Gives one orthogonal vector of a vector

    Given `vector` find one orthogonal vector

    Parameters
    ----------
    vector : Base.Vector

    Returns
    -------
    orthogonal : Base.Vector
    """
    min_pos = np.argmin([abs(vector[0]), abs(vector[1]), abs(vector[2])])
    if min_pos == 0:
        orthogonal = Base.Vector(0, vector[2], -vector[1])
    elif min_pos == 1:
        orthogonal = Base.Vector(vector[2], 0, -vector[0])
    else:
        orthogonal = Base.Vector(vector[1], -vector[0], 0)
    return orthogonal.normalize()


def correct_normal(normal, incident):
    """Corrects a vector so that is in a given half plane

    Parameters
    ----------
    normal : Base.Vector
    incident : Base.Vector

    Returns
    -------

    """
    if normal.dot(incident) > 0:
        return normal * (-1)
    else:
        return normal


def normalize(vector):
    """Normalizes a vector"""
    if vector.Length < EPSILON:
        vector = vector * INF
    return vector.normalize()


def arccos(x):
    """Safe modification of arccos"""
    assert abs(x) < 1 + EPSILON
    if abs(x) < 1 - EPSILON:
        return np.arccos(x)
    if x > 0:
        return 0
    return np.pi


def projection_on_vector(u, v):
    """Compute the projection of u on <v>"""
    return (u.dot(v)/v.dot(v))*v


def projection_on_orthogonal_of_vector(u, v):
    """Compute the projection of u on the subspace orthogonal to <v>"""
    return u - projection_on_vector(u, v)


def area_of_triangle(vertices):
    """Compute the area of the triangle with given vertices"""
    p, q, r = vertices
    pq = q-p
    pr = r-p
    v = pq.cross(pr)
    return 0.5 * abs(v.Length)


def random_point_of_triangle(vertices):
    """Compute a random point of the triangle with given vertices"""
    p, q, r = vertices
    pq = q-p
    pr = r-p
    while True:
        x = random.random()
        y = random.random()
        if x + y <= 1:
            return p + pq*x + pr*y


"""Module otsun.optics

Implementation of optical effects on rays
"""

from FreeCAD import Base
import numpy as np
from .math import myrandom, one_orthogonal_vector, correct_normal, \
    parallel_orthogonal_components, normalize
from enum import Enum
from numpy.lib.scimath import sqrt
from autologging import traced
import logging
logger = logging.getLogger(__name__)

# from .materials import Material, vacuum_medium

EPSILON = 1E-6
# Tolerance for considering equal to zero
INF = 1E20


class Phenomenon(Enum):
    """
    Enum for optical phenomena.

    This enum is used to describe which optical phenomenon affected a ray
    """
    JUST_STARTED = 0
    REFLEXION = 1
    REFRACTION = 2
    ABSORPTION = 3
    TRANSMITTANCE = 4
    ENERGY_ABSORBED = 5


class OpticalState(object):
    """Optical state of a ray.

    The OpticalState class gathers together information about the optical
    state of a ray.

    Parameters
    ----------
    polarization : Base.Vector
        polarization vector of the ray
    direction : Base.Vector
        direction vector of the ray
    phenomenon : Phenomenon
        last phenomenon that the ray experimented
    material : Material
        Material where the ray is located
    extra_data : dict
        Dictionary where materials can put extra data
    """

    def __init__(self, polarization, direction, phenomenon, material=None, extra_data=None):
        self.polarization = polarization
        self.direction = direction
        self.phenomenon = phenomenon
        self.material = material  # TODO: Set solid
        if extra_data is None:
            extra_data = {}
        self.extra_data = extra_data

    def __str__(self):
        return "Dir.: %s, Pol.: %s, Phen.: %s" % (
            self.direction, self.polarization, self.phenomenon
        )

    def apply_single_gaussian_dispersion(self, normal, sigma_1):
        """
        Apply a single gaussian dispersion to the optical state

        Parameters
        ----------
        normal : Base.Vector
            normal vector of the surface at the point of incidence
        sigma_1 : float
            dispersion coefficient
         """
        rad = np.pi / 180.0
        u = myrandom()
        theta = (-2. * sigma_1 ** 2. * np.log(u)) ** 0.5 / 1000.0 / rad
        phi = 360. * myrandom()
        new_direction = dispersion_from_main_direction(self.direction, 2 * theta, phi)
        new_polarization_vector = dispersion_polarization(self.direction, self.polarization, 2 * theta, phi)
        self.polarization = new_polarization_vector
        self.direction = new_direction

    def apply_double_gaussian_dispersion(self, normal, sigma_1, sigma_2, k):
        """
        Apply a double gaussian dispersion to the state


        Parameters
        ----------
        normal : Base.Vector
            normal vector of the surface at the point of incidence
        sigma_1 : float
            dispersion coefficient (first case)
        sigma_2 : float
            dispersion coefficient (second case)
        k : float
            threshold for randomly applying first or second case
        """
        rad = np.pi / 180.0
        k_ran = myrandom()
        u = myrandom()
        if k_ran < k:
            theta = (-2. * sigma_1 ** 2. * np.log(u)) ** 0.5 / 1000.0 / rad
        else:
            theta = (-2. * sigma_2 ** 2. * np.log(u)) ** 0.5 / 1000.0 / rad
        phi = 360. * myrandom()
        new_direction = dispersion_from_main_direction(self.direction, 2 * theta, phi)
        new_polarization_vector = dispersion_polarization(self.direction, self.polarization, 2 * theta, phi)
        self.polarization = new_polarization_vector
        self.direction = new_direction

    def apply_dispersion(self, properties, normal_vector):
        if properties.get('sigma_1', None):
            sigma_1 = properties['sigma_1']
            if properties.get('sigma_2', None):
                sigma_2 = properties['sigma_2']
                k = properties.get('k', None) or 0.5
                self.apply_double_gaussian_dispersion(
                    normal_vector, sigma_1, sigma_2, k)
            else:
                self.apply_single_gaussian_dispersion(normal_vector, sigma_1)


# ---
# Helper functions for reflection and refraction
# ---


@traced(logger)
def simple_reflection(incident, normal):
    """
    Returns the reflection of the vector `incident` on a surface orthogonal to `normal`
    Parameters
    ----------
    incident : Base.Vector
    normal : Base.Vector

    Returns
    -------
    Base.Vector
    """
    normal = correct_normal(normal, incident)
    c1 = - normal.dot(incident)
    reflected = incident + normal * 2.0 * c1
    return reflected.normalize()


@traced(logger)
def outgoing_polarization(incident_vector, incident_polarization, outgoing_vector,
                          perpendicularly_polarized=False):
    if perpendicularly_polarized:
        return incident_polarization
    angle = incident_vector.getAngle(outgoing_vector)
    try:
        normal = incident_vector.cross(outgoing_vector)
        normal.normalize()
    except Base.FreeCADError:
        # we assume now that the two vectors are collinear
        normal = one_orthogonal_vector(incident_vector)
    rotation = Base.Rotation(normal, angle)
    return rotation.multVec(incident_polarization)


@traced(logger)
def reflection(incident, normal_vector, polarization_vector):
    """
    Returns the OpticalState of a ray with direction `incident` and polarized in `polarization_vector` after
    being reflected on a surface orthogonal to `normal_vector`

    Parameters
    ----------
    incident : Base.Vector
        vector incident
    normal_vector : Base.Vector
        vector normal to the surface
    polarization_vector: Base.Vector
        Polarization vector of the ray

    Returns
    -------
    OpticalState
        optical state of the reflected ray
    """
    normal = correct_normal(normal_vector, incident)
    reflected = simple_reflection(incident, normal)
    polarization_vector_out = outgoing_polarization(incident, polarization_vector, reflected)
    return OpticalState(polarization_vector_out, reflected, Phenomenon.REFLEXION)  # TODO: Set solid


@traced(logger)
def total_lambertian_reflection(incident, normal_vector, polarization_vector):
    """
    Returns the OpticalState of a ray with direction `incident` and polarized in `polarization_vector` after
    being reflected on a surface orthogonal to `normal_vector` with isotropic lambertian properties.

    Parameters
    ----------
    polarization_vector
    incident : Base.Vector
        vector incident
    normal_vector : Base.Vector
        vector normal to the surface

    Returns
    -------
    OpticalState
        optical state of the reflected ray
    """
    normal = correct_normal(normal_vector, incident)
    dot = - 1.0
    while dot <= 0.01:
        random_vector = Base.Vector(myrandom() - 0.5, myrandom() - 0.5, myrandom() - 0.5)
        if random_vector.Length < 1:
            random_vector.normalize()
            dot = normal.dot(random_vector)
    new_direction = random_vector
    random_polarization_vector = random_polarization(new_direction)
    return OpticalState(random_polarization_vector,
                        new_direction, Phenomenon.REFLEXION)  # TODO: Set solid


@traced(logger)
def cosine_lambertian_reflection(incident, normal_vector, polarization_vector):
    """
    Returns the OpticalState of a ray with direction `incident` and polarized in `polarization_vector` after
    being reflected on a surface orthogonal to `normal_vector` with cosine-law lambertian properties.

    Parameters
    ----------
    incident : Base.Vector
        vector incident
    normal_vector : Base.Vector
        vector normal to the surface

    Returns
    -------
    OpticalState
        optical state of the reflected ray
    """
    normal = correct_normal(normal_vector, incident)
    theta = np.arcsin(myrandom()) * 180 / np.pi
    phi = 360. * myrandom()
    new_direction = dispersion_from_main_direction(normal, theta, phi)
    new_polarization_vector = dispersion_polarization(normal, polarization_vector, theta, phi)
    return OpticalState(new_polarization_vector,
                        new_direction, Phenomenon.REFLEXION)  # TODO: Set solid


@traced(logger)
def reflection_hub(incident, normal_vector, polarization_vector,
                   lambertian_weight=0, lambertian_kind=None):
    if myrandom() >= lambertian_weight:
        # no lambertian phenomenon
        return reflection(incident, normal_vector, polarization_vector)
    elif lambertian_kind == "Total":
        return total_lambertian_reflection(incident, normal_vector, polarization_vector)
    elif lambertian_kind == "Cosine":
        return cosine_lambertian_reflection(incident, normal_vector, polarization_vector)
    else:
        raise Exception(f"Lambertian kind {lambertian_kind} not implemented")


@traced(logger)
def refraction(incident, normal_vector, n1, n2, polarization_vector,
               lambertian_weight=0, lambertian_kind=None):
    """
    Returns the OpticalState of a ray that hits the surface (orthogonal to `normal_vector`) that is the interface
    between mediums of refractive indices `n1` and `n2`, with direction `incident` and polarized in
    `polarization_vector`. In case of reflection, the surface acts as a lambertian layer with probability
    `lambertian_weight` and of kind `lambertian_kind` (either "Total" or "Cosine").


    Parameters
    ----------
    incident : Base.Vector
        direction vector of the incident ray	
    normal_vector: Base.Vector
        normal vector of the surface at the point of incidence
    n1: complex
        complex refractive index where ray is currently traveling
    n2: complex
        complex refractive index of nearby material
    polarization_vector: Base.Vector
        Polarization vector of the ray
    lambertian_weight: Float
        Probability that the surface acts as Lambertian (0=never; 1=always)
    lambertian_kind: str
    Returns
    -------
    OpticalState
    """
    # Fresnel equations. Oblique incident in absorbing media.
    # See Chapter 2 of the book Thin-films optical filters
    # See Fresnel equations in WikipediA
    normal = correct_normal(normal_vector, incident)
    r = n1 / n2
    c1 = - normal.dot(incident)
    # cos (incident_angle)
    c2sq = 1.0 - r * r * (1.0 - c1 * c1)
    # cos (refracted_angle) ** 2
    if c2sq.real < 0:
        # total internal reflection
        return reflection_hub(incident, normal, polarization_vector,
                              lambertian_weight, lambertian_kind)
    c2 = sqrt(c2sq)
    # cos (refracted_angle)
    parallel_v, perpendicular_v, normal_parallel_plane = parallel_orthogonal_components(polarization_vector, incident,
                                                                                        normal)
    # parallel and perpendicular components of polarization vector and orthogonal vector of the parallel plane
    ref_per = perpendicular_v.Length ** 2.0 / polarization_vector.Length ** 2.0
    # weight of perpendicular component: 0 < ref_per < 1
    # We decide the polarization projection onto the parallel / perpendicular plane
    if myrandom() < ref_per:
        # reflectance for s-polarized (perpendicular) light
        a = (n1 * c1 - n2 * c2) / (n1 * c1 + n2 * c2)
        reflectance = a * a.conjugate()
        perpendicular_polarized = True
        polarization_vector = normalize(perpendicular_v)
    else:
        # reflectance for p-polarized (parallel) light
        a = (n1 * c2 - n2 * c1) / (n1 * c2 + n2 * c1)
        reflectance = a * a.conjugate()
        perpendicular_polarized = False
        polarization_vector = normalize(parallel_v)
    # The ray can be reflected or refracted 
    if myrandom() < reflectance.real:
        return reflection_hub(incident, normal, polarization_vector,
                              lambertian_weight, lambertian_kind)
    else:
        # ray refracted: computing the refracted direction
        if c2.real > 1:
            # avoiding invalid solutions for metallic materials
            c2 = 1
        refracted_direction = incident * r.real + normal * (r.real * c1 - c2.real)
        refracted_direction.normalize()
        polarization_vector_out = outgoing_polarization(incident, polarization_vector,
                                                        refracted_direction, perpendicular_polarized)
        return OpticalState(polarization_vector_out, refracted_direction,
                            Phenomenon.REFRACTION)  # TODO: Set solid


@traced(logger)
def shure_refraction(incident, normal_vector, n1, n2, polarization_vector,
                     lambertian_weight=0, lambertian_kind=None):
    """Implementation of Snell's law of refraction

    Parameters
    ----------
    incident : Base.Vector
        direction vector of the incident ray
    normal_vector: Base.Vector
        normal vector of the surface at the point of incidence
    n1: complex
        complex refractive index where ray is currently traveling
    n2: complex
        complex refractive index of nearby material
    polarization_vector: Base.Vector
        Polarization vector of the ray
    lambertian_weight: Float
        Probability that the surface acts as lambertian

    Returns
    -------
    OpticalState
    """
    # TODO: document

    normal = correct_normal(normal_vector, incident)
    r = n1 / n2
    c1 = - normal.dot(incident)
    # cos (incident_angle)
    c2sq = 1.0 - r * r * (1.0 - c1 * c1)
    # cos (refracted_angle) ** 2
    if c2sq.real < 0:
        # total internal reflection
        return reflection_hub(incident, normal, polarization_vector,
                              lambertian_weight, lambertian_kind)
    c2 = sqrt(c2sq)
    # cos (refracted_angle)
    if c2.real > 1:
        # avoiding invalid solutions for metallic materials
        c2 = 1
    #parallel_v, perpendicular_v, normal_parallel_plane = parallel_orthogonal_components(polarization_vector, incident,
    #                                                                                    normal)
    # parallel and perpendicular components of polarization vector and orthogonal vector of the parallel plane
    # ray refracted: computing the refracted direction
    refracted_direction = incident * r.real + normal * (r.real * c1 - c2.real)
    refracted_direction.normalize()
    polarization_vector_out = outgoing_polarization(incident, polarization_vector, refracted_direction)
    return OpticalState(polarization_vector_out, refracted_direction,
                        Phenomenon.REFRACTION)  # TODO: Set solid


# ---
# Helper function for dispersions and polarization vector
# ---
@traced(logger)
def dispersion_from_main_direction(main_direction, theta, phi):
    """
    Computes dispersion from the main direction in terms of angles theta and phi
    """
    v = main_direction
    v_p = Base.Vector(v[1], -v[0], 0)
    if v_p == Base.Vector(0, 0, 0):
        # to avoid null vector at mynormal and incident parallel vectors
        v_p = Base.Vector(1, 0, 0)
    v_p.normalize()
    rotation_1 = Base.Rotation(v_p, theta)
    new_v1 = rotation_1.multVec(v)
    axis_2 = main_direction
    rotation_2 = Base.Rotation(axis_2, phi)
    new_v2 = rotation_2.multVec(new_v1)
    return new_v2


@traced(logger)
def dispersion_polarization(main_direction, polarization_vector, theta, phi):
    """
    Computes dispersion of polarization vector in terms of angles theta and phi
    """
    v = main_direction
    v_p = Base.Vector(v[1], -v[0], 0)
    if v_p == Base.Vector(0, 0, 0):
        # to avoid null vector at mynormal and incident parallel vectors
        v_p = Base.Vector(1, 0, 0)
    v_p.normalize()
    rotation_1 = Base.Rotation(v_p, theta)
    new_v1 = rotation_1.multVec(polarization_vector)
    axis_2 = main_direction
    rotation_2 = Base.Rotation(axis_2, phi)
    new_v2 = rotation_2.multVec(new_v1)
    return new_v2


@traced(logger)
def random_polarization(direction):
    """
    Returns a random polarization vector, orthogonal to `direction`
    """
    orthogonal_vector = one_orthogonal_vector(direction)
    phi = 360.0 * myrandom()
    rotation = Base.Rotation(direction, phi)
    random_polarization_v = rotation.multVec(orthogonal_vector)
    return random_polarization_v


# ---
# Helper function for reflectance depending on the wavelength for coatings layers.
# ---

def _round_or_floor_ceil(x):
    xround = int(round(x))
    if abs(x - xround) < 1E-4:
        return [xround]
    xfloor = int(np.floor(x))
    return [xfloor, xfloor + 1]


@traced(logger)
def matrix_reflectance(data_material):
    """
    Computes the matrix of reflectances of a material

    data_material: wavelenth in nm, angle in deg., reflectance s-polarized (perpendicular),
    reflectance p-polarized (parallel)
    the values should be in the corresponding order columns with constants steps
    We generate a matrix for the interpolation depending on the x=angle and
    y=wavelength values (lambda values)

    Parameters
    ----------
    data_material
    """
    steps = np.diff(data_material[:, 1])
    try:
        delta_a = min(steps[steps > 0.0])
    except ValueError:
        delta_a = 1E-4
    min_a = min(data_material[:, 1])
    steps = np.diff(data_material[:, 0])
    try:
        delta_w = min(steps[steps > 0.0])
    except ValueError:
        delta_w = 1E-4
    min_w = min(data_material[:, 0])
    data_dict = {}
    for row in data_material:
        data_dict[(int(round((row[0] - min_w) / delta_w)),
                   int(round((row[1] - min_a) / delta_a)))] = row

    #    @memoize
    # @lru_cache(maxsize=None)
    def internal_matrix_reflectance(x, y):
        index_a = (x - min_a) / delta_a
        index_w = (y - min_w) / delta_w
        index_a_ints = _round_or_floor_ceil(index_a)
        index_w_ints = _round_or_floor_ceil(index_w)
        rows = []
        for i_a in index_a_ints:
            for i_w in index_w_ints:
                try:
                    rows.append(data_dict[(i_w, i_a)])
                except KeyError:
                    pass
        return rows

    return internal_matrix_reflectance


@traced(logger)
def calculate_reflectance(m_reflectance, angle, wavelength):
    """Compute perpendicular and parallel components of reflectance

    Interpolates the value of the perperdicular and parallel reflectance from
    the matrix of values, depending on the angle and the wavelength

    Parameters
    ----------
    m_reflectance : float or tuple of list of floats
    angle : float
    wavelength : float

    Returns
    -------
    tuple of float or tuple of np.complex
    """
    if len(m_reflectance) == 0:  # wavelength experiments out of range of the data_material
        return 0.0, 0.0
    r_matrix = np.mat(m_reflectance)
    if len(r_matrix) == 1:
        # interpolation is not needed
        return r_matrix[0, 2], r_matrix[0, 3]

    if len(r_matrix) == 2:  # simple interpolation
        if r_matrix[0, 0] == r_matrix[1, 0]:
            # identical wavelengths; hence angle interpolation (column 1)
            x_values = [r_matrix[0, 1], r_matrix[1, 1]]
            y_values = [r_matrix[0, 2], r_matrix[1, 2]]
            r_per = np.interp(angle, x_values, y_values)

            x_values = [r_matrix[0, 1], r_matrix[1, 1]]
            y_values = [r_matrix[0, 3], r_matrix[1, 3]]
            r_par = np.interp(angle, x_values, y_values)
            return r_per, r_par

        if r_matrix[0, 1] == r_matrix[1, 1]:
            # identical angles; hence wavelength interpolation (column 0)
            x_values = [r_matrix[0, 0], r_matrix[1, 0]]
            y_values = [r_matrix[0, 2], r_matrix[1, 2]]
            r_per = np.interp(wavelength, x_values, y_values)

            x_values = [r_matrix[0, 0], r_matrix[1, 0]]
            y_values = [r_matrix[0, 3], r_matrix[1, 3]]
            r_par = np.interp(wavelength, x_values, y_values)
            return r_per, r_par

    if len(r_matrix) == 4:
        # double interpolation
        x_values = [r_matrix[0, 1], r_matrix[1, 1]]
        y_values = [r_matrix[0, 2], r_matrix[1, 2]]
        r1 = np.interp(angle, x_values, y_values)
        x_values = [r_matrix[2, 1], r_matrix[3, 1]]
        y_values = [r_matrix[2, 2], r_matrix[3, 2]]
        r2 = np.interp(angle, x_values, y_values)
        x_values = [r_matrix[0, 0], r_matrix[3, 0]]
        y_values = [r1, r2]
        r_per = np.interp(wavelength, x_values, y_values)

        x_values = [r_matrix[0, 1], r_matrix[1, 1]]
        y_values = [r_matrix[0, 3], r_matrix[1, 3]]
        r1 = np.interp(angle, x_values, y_values)
        x_values = [r_matrix[2, 1], r_matrix[3, 1]]
        y_values = [r_matrix[2, 3], r_matrix[3, 3]]
        r2 = np.interp(angle, x_values, y_values)
        x_values = [r_matrix[0, 0], r_matrix[3, 0]]
        y_values = [r1, r2]
        r_par = np.interp(wavelength, x_values, y_values)
        return r_per, r_par



"""Module otsun.experiments for creating experiments.

The module defines a class `Experiment` that deals with the setup and
run of experiments.
"""

import numpy as np
import logging
logger = logging.getLogger(__name__)

class Experiment:
    """
    Sets up and runs and experiment in a given scene with a given light source.

    An Experiment is formed by a scene and a ligh_source. Rays are emitted from light_source
    and interfere with the scene until they are absorbed or scape from the scene. If they are
    absorbed, the energies are computed and stored.

    Parameters
    ----------
    scene : otsun.Scene
        Scene of the experiment
    light_source : otsun.LightSource
        Source of the rays to emit
    number_of_rays : int
        Number of rays to emit in the experiment
    show_in_doc : App.Document
        FreeCAD document where to plot the rays, or None if plotting is not desired

    Attributes
    ----------
    wavelengths: List of float
        List of wavelengths of emitted rays
    captured_energy_Th : float
        Total energy of rays that got absorbed
    captured_energy_PV : float
        Total energy of rays that fell in a PV
    Th_energy :
    Th_wavelength : List of float
        List of wavelengths of rays that got absorbed
    PV_energy : List of floats
        List of energies of rays that fell in a PV
    PV_wavelength : List of float
        List of wavelengths of rays that fell in a PV
    PV_values : List of tuples of floats
        List of PV_values of all emitted rays that fell in a PV
    points_absorber_Th : List of tuple of floats
        List with data (energy, location,...) for each ray that got absorbed
    """

    def __init__(self, scene, light_source, number_of_rays, show_in_doc=None):
        self.scene = scene
        self.light_source = light_source
        if show_in_doc:
            self.light_source.emitting_region.add_to_document(show_in_doc)
        self.number_of_rays = number_of_rays
        self.wavelengths = []
        self.captured_energy_Th = 0
        self.captured_energy_PV = 0
        self.Th_energy = []
        self.Th_wavelength = []
        self.PV_energy = []
        self.PV_wavelength = []
        self.PV_values = []
        self.points_absorber_Th = []

    def run(self, show_in_doc=None):
        """Runs the experiment and plots the rays in the document specified (if any)"""
        for n in np.arange(0, self.number_of_rays, 1):
            logger.info(f"Emitting ray {n}")
            ray = self.light_source.emit_ray()
            ray.run()
            self.wavelengths.append(ray.wavelength)
            if show_in_doc:
                ray.add_to_document(show_in_doc)
            if ray.Th_absorbed:
                self.captured_energy_Th += ray.energy
                self.Th_energy.append(ray.energy)
                self.Th_wavelength.append(ray.wavelength)
                self.points_absorber_Th.append((ray.energy,
                                                ray.points[-1].x, ray.points[-1].y, ray.points[-1].z,
                                                ray.points[-2].x, ray.points[-2].y, ray.points[-2].z,
                                                ray.last_normal.x, ray.last_normal.y, ray.last_normal.z,
												ray.wavelength))
            else:
                self.Th_energy.append(0.0)
                # TODO: Review... ray.wavelength always added to Th_wavelength
                # Hence always Th_wavelength == wavelenghts
                self.Th_wavelength.append(ray.wavelength)
            if ray.PV_absorbed:
                PV_energy_absorbed = np.sum(ray.PV_absorbed)
                self.captured_energy_PV += PV_energy_absorbed
                self.PV_energy.append(PV_energy_absorbed)
                self.PV_wavelength.append(ray.wavelength)
                length = len(ray.PV_values)
                if length > 0:
                    for z in np.arange(0, length, 1):
                        # TODO: Review... no ho entenc (pq no afegir directament tot?) Ramon please check
                        self.PV_values.append(ray.PV_values[z])
            else:
                self.PV_energy.append(0.0)
                # TODO: Review... ray.wavelength always added to PV_wavelength
                self.PV_wavelength.append(ray.wavelength)



"""Module otsun.scene for the modelization optical systems

The module defines the class `Scene` that models the elements in an optical system
"""

from .materials import Material, VolumeMaterial, SurfaceMaterial, TwoLayerMaterial, MaterialNotFoundError
from .math import correct_normal
from autologging import traced
import logging
logger = logging.getLogger(__name__)

EPSILON = 1E-6

@traced(logger)
class Scene:
    """
    Class used to define the Scene. It encodes all the objects
    that interfere with light rays.
    """

    def __init__(self, objects, extra_data=None):
        self.objects = objects
        self.faces = []  # All the faces in the Scene
        self.solids = []  # All the solids in the Scene
        self.name_of_solid = {}
        self.bb_of_solid = {}  # Bounding box of solid
        self.materials = {}  # Assign the materials to objects
        # self.boundaries = {}  # Assign boundary faces to solids
        # self.sum_energy = 0.0 # Energy of the Scene
        self.epsilon = EPSILON # Tolerance for solid containment # 2 nm.
        self.boundbox = None
        self.element_object_dict = {}

        if extra_data is None:
            self.extra_data = {}
        else:
            self.extra_data = extra_data
        # Format for extra_data:
        # extra_data['vacuum_material']: VolumeMaterial to be considered after 1st intersection
        if 'vacuum_material' in self.extra_data:
            vacuum_material_name = self.extra_data['vacuum_material'][0]
            material = Material.by_name.get(vacuum_material_name, None)
            if not material:
                logger.warning("Material %s not found for vacuum", vacuum_material_name)
            else:
                self.extra_data['vacuum_material'] = material
        if 'axis_deviation' in self.extra_data:
            pairs_list = self.extra_data['axis_deviation'][:]
            self.extra_data['axis_deviation'] = {}
            iterator = iter(pairs_list)
            pairs_iterator = zip(iterator, iterator)
            for axis, sigma in pairs_iterator:
                self.extra_data['axis_deviation'][axis] = sigma
        # extra_data['axis_deviation']: Dictionary where each key is the axis in the scene and the value its std dev

        for obj in objects:
            # noinspection PyNoneFunctionAssignment
            logger.debug(f"loading object {obj.Label}")
            try:
                material = Material.get_from_label(obj.Label)
            except MaterialNotFoundError as e:
                logger.warning(f"{e}")
                continue
            if not material:
                logger.debug(f"Object {obj.Label} does not have a material")
                continue
            shape = obj.Shape
            faces = shape.Faces
            solids = shape.Solids
            if isinstance(material, VolumeMaterial):
                if not solids:
                    logger.warning(f"Volume material {material.name} associated to object {obj.Label} without solids")
                for solid in solids:
                    logger.debug(f"Assigning material {material.name} to solid {solid} in {obj.Label}")
                    self.name_of_solid[solid] = obj.Label
                    self.materials[id(solid)] = material
                    self.element_object_dict[solid] = obj
                    self.bb_of_solid[solid] = solid.BoundBox
                self.solids.extend(solids)
                self.faces.extend(faces)
            elif isinstance(material, SurfaceMaterial) or isinstance(material, TwoLayerMaterial):
                if solids:
                    logger.warning(f"Surface material {material.name} associated to object {obj.Label} with solids")
                if not faces:
                    logger.warning(f"Surface material {material.name} associated to object {obj.Label} without faces")
                for face in faces:
                    logger.debug(f"Assigning material {material.name} to face {face} in {obj.Label}")
                    self.materials[id(face)] = material
                    self.element_object_dict[face] = obj
                self.faces.extend(faces)
            else:
                logger.warning(f"Material {material.name} associated to {obj.Label} is not Surface or Volume material")
                continue
            # solids = obj.Shape.Solids
            # faces = obj.Shape.Faces
            # if solids and isinstance(material, VolumeMaterial):  # Object is a solid
            #     for solid in solids:
            #         logger.debug(f"Assigning material {material.name} to solid {solid} in {obj.Label}")
            #         self.name_of_solid[solid] = obj.Label
            #         self.materials[solid] = material
            #         self.element_object_dict[solid] = obj
            # else:  # Object is made of faces
            #     for face in faces:
            #         logger.debug(f"Assigning material {material.name} to face {face} in {obj.Label}")
            #         self.materials[face] = material
            #         self.element_object_dict[face] = obj
            # self.solids.extend(solids)
            # self.faces.extend(faces)
            # # self.materials[obj] = material ## Cal?
            if not self.boundbox:
                self.boundbox = obj.Shape.BoundBox
            else:
                self.boundbox.add(obj.Shape.BoundBox)

        # self.remove_duplicate_faces()

        self.diameter = self.boundbox.DiagonalLength

    def recompute_boundbox(self):
        """
        Recomputes the boundbox, so that all objects are contained in it
        """
        boundbox = None
        for elem in self.solids:
            if boundbox is None:
                boundbox = elem.BoundBox
            else:
                boundbox.add(elem.BoundBox)
        for elem in self.faces:
            if boundbox is None:
                boundbox = elem.BoundBox
            else:
                boundbox.add(elem.BoundBox)
        self.boundbox = boundbox

    def remove_duplicate_faces(self):
        """
        Removes redundant faces, so that the computation of next_intersection does not find duplicated points.
        """
        logger.debug("Removing duplicate faces")
        faces_with_material = [face for face in self.faces if
                               id(face) in self.materials]
        faces_no_material = [face for face in self.faces if
                             id(face) not in self.materials]
        if not faces_no_material:
            logger.debug("Done Removing duplicate faces (nothing to do)")
            return
        complex_no_material = faces_no_material[0]
        for (i,face) in enumerate(faces_no_material[1:]):
            logger.debug(f"Fusing face {i} of {len(faces_no_material)-1}")
            complex_no_material = complex_no_material.fuse(face)
        for (i,face) in enumerate(faces_with_material):
            logger.debug(f"Cutting face {i} of {len(faces_with_material) - 1}")
            complex_no_material = complex_no_material.cut(face)
        self.faces = faces_with_material
        self.faces.extend(complex_no_material.Faces)
        logger.debug("Done Removing duplicate faces")

    def solid_at_point(self, point):
        """
        Returns the solid that a point is inside.
        """
        for solid in self.solids:
            bb = self.bb_of_solid[solid]
            if bb.isInside(point):
                if solid.isInside(point, self.epsilon, False):
                    return solid
        return None

    def face_at_point(self, point):
        """
        Returns the face that a point is inside.
        """
        for face in self.faces:
            if face.isInside(point, self.epsilon, True):
                return face
        return None

    def next_solid_at_point_in_direction(self, point, normal, direction):
        """
        Returns the next solid found in the given direction from a given point
        """
        external_normal = correct_normal(normal, direction)
        point_plus_epsilon = point + external_normal * (-2) * self.epsilon
        return self.solid_at_point(point_plus_epsilon)
